import pytest
from django.contrib.auth import get_user_model

from country.factories import CountryFactory
from region.factories import RegionFactory
from user.factories.account import UserAccountFactory
from user.factories.address import UserAddressFactory
from user.models.address import UserAddress

User = get_user_model()


@pytest.mark.django_db
class TestUserAddressManager:
    def setup_method(self):
        self.user = UserAccountFactory()
        self.other_user = UserAccountFactory()

    def test_get_user_addresses_empty(self):
        addresses = UserAddress.objects.get_user_addresses(self.user)

        assert addresses.count() == 0
        assert list(addresses) == []

    def test_get_user_addresses_single(self):
        address = UserAddressFactory(user=self.user, title="Home")

        addresses = UserAddress.objects.get_user_addresses(self.user)

        assert addresses.count() == 1
        assert address in addresses

    def test_get_user_addresses_multiple(self):
        address1 = UserAddressFactory(
            user=self.user, title="Home", is_main=True
        )
        address2 = UserAddressFactory(
            user=self.user, title="Work", is_main=False
        )
        address3 = UserAddressFactory(
            user=self.user, title="Secondary", is_main=False
        )

        UserAddressFactory(user=self.other_user, title="Other Home")

        addresses = UserAddress.objects.get_user_addresses(self.user)

        assert addresses.count() == 3
        assert address1 in addresses
        assert address2 in addresses
        assert address3 in addresses

    def test_get_user_addresses_ordering(self):
        work_address = UserAddressFactory(
            user=self.user, title="Work", is_main=False
        )
        home_address = UserAddressFactory(
            user=self.user, title="Home", is_main=True
        )
        secondary_address = UserAddressFactory(
            user=self.user, title="Secondary", is_main=False
        )

        addresses = list(UserAddress.objects.get_user_addresses(self.user))

        assert len(addresses) == 3
        assert addresses[0] == home_address
        assert addresses[1] == secondary_address
        assert addresses[2] == work_address

    def test_get_main_address_exists(self):
        regular_address = UserAddressFactory(
            user=self.user, title="Work", is_main=False
        )
        main_address = UserAddressFactory(
            user=self.user, title="Home", is_main=True
        )

        result = UserAddress.objects.get_main_address(self.user)

        assert result == main_address
        assert result != regular_address

    def test_get_main_address_none_exists(self):
        UserAddressFactory(user=self.user, title="Work", is_main=False)
        UserAddressFactory(user=self.user, title="Home", is_main=False)

        result = UserAddress.objects.get_main_address(self.user)

        assert result is None

    def test_get_main_address_no_addresses(self):
        result = UserAddress.objects.get_main_address(self.user)

        assert result is None

    def test_get_main_address_user_isolation(self):
        _ = UserAddressFactory(
            user=self.other_user, title="Other Home", is_main=True
        )

        UserAddressFactory(user=self.user, title="Our Work", is_main=False)

        result = UserAddress.objects.get_main_address(self.user)

        assert result is None

    def test_manager_inheritance(self):
        assert isinstance(UserAddress.objects, UserAddress.objects.__class__)
        assert hasattr(UserAddress.objects, "get_user_addresses")
        assert hasattr(UserAddress.objects, "get_main_address")

    def test_get_user_addresses_with_related_data(self):
        country = CountryFactory()
        region = RegionFactory(country=country)

        address1 = UserAddressFactory(
            user=self.user,
            title="International",
            country=country,
            region=region,
            is_main=True,
        )
        address2 = UserAddressFactory(
            user=self.user,
            title="Local",
            country=None,
            region=None,
            is_main=False,
        )

        addresses = UserAddress.objects.get_user_addresses(self.user)

        assert addresses.count() == 2
        assert address1 in addresses
        assert address2 in addresses

        ordered_addresses = list(addresses)
        assert ordered_addresses[0] == address1
        assert ordered_addresses[1] == address2

    def test_main_address_uniqueness_enforcement(self):
        main1 = UserAddressFactory(
            user=self.user, title="First Main", is_main=True
        )

        main2 = UserAddressFactory(
            user=self.user, title="Second Main", is_main=True
        )

        main1.refresh_from_db()

        assert not main1.is_main
        assert main2.is_main

        current_main = UserAddress.objects.get_main_address(self.user)
        assert current_main == main2

    def test_manager_with_different_users(self):
        user1_home = UserAddressFactory(
            user=self.user, title="User1 Home", is_main=True
        )
        user1_work = UserAddressFactory(
            user=self.user, title="User1 Work", is_main=False
        )

        user2_home = UserAddressFactory(
            user=self.other_user, title="User2 Home", is_main=True
        )
        user2_work = UserAddressFactory(
            user=self.other_user, title="User2 Work", is_main=False
        )

        user1_addresses = UserAddress.objects.get_user_addresses(self.user)
        user1_main = UserAddress.objects.get_main_address(self.user)

        assert user1_addresses.count() == 2
        assert user1_home in user1_addresses
        assert user1_work in user1_addresses
        assert user1_main == user1_home

        user2_addresses = UserAddress.objects.get_user_addresses(
            self.other_user
        )
        user2_main = UserAddress.objects.get_main_address(self.other_user)

        assert user2_addresses.count() == 2
        assert user2_home in user2_addresses
        assert user2_work in user2_addresses
        assert user2_main == user2_home

    def test_queryset_methods_are_chainable(self):
        UserAddressFactory(
            user=self.user, title="Home", city="New York", is_main=True
        )
        UserAddressFactory(
            user=self.user, title="Work", city="Boston", is_main=False
        )

        ny_addresses = UserAddress.objects.get_user_addresses(self.user).filter(
            city="New York"
        )
        assert ny_addresses.count() == 1

        addresses_by_city = UserAddress.objects.get_user_addresses(
            self.user
        ).order_by("city")
        cities = [addr.city for addr in addresses_by_city]
        assert cities == ["Boston", "New York"]



import django.dispatch
from asgiref.sync import sync_to_async
from django.conf import settings
from django.dispatch import receiver
from simple_history.signals import post_create_historical_record

from notification.enum import NotificationKindEnum
from notification.models.notification import Notification
from notification.models.user import NotificationUser
from product.models.favourite import ProductFavourite
from product.models.product import Product

languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]

product_price_lowered = django.dispatch.Signal()
product_price_increased = django.dispatch.Signal()


@receiver(post_create_historical_record)
def post_create_historical_record_callback(
    sender, instance, history_instance, **kwargs
):
    prev_record = getattr(history_instance, "prev_record", None)
    if prev_record is None:
        return

    old_price = prev_record.price.amount
    new_price = instance.price.amount

    if old_price > new_price:
        product_price_lowered.send(
            sender=Product,
            instance=instance,
            old_price=old_price,
            new_price=new_price,
        )
    elif old_price < new_price:
        product_price_increased.send(
            sender=Product,
            instance=instance,
            old_price=old_price,
            new_price=new_price,
        )


@receiver(product_price_lowered)
async def notify_product_price_lowered(
    sender, instance, old_price, new_price, **kwargs
):
    favorite_users = await sync_to_async(
        lambda: list(
            ProductFavourite.objects.filter(product=instance).select_related(
                "user"
            )
        )
    )()

    product_url = (
        f"{settings.NUXT_BASE_URL}/products/{instance.id}/{instance.slug}"
    )

    async def get_instance_name():
        return await sync_to_async(
            lambda: instance.safe_translation_getter("name", any_language=True)
            or f"Product {instance.slug or instance.id}",
            thread_sensitive=True,
        )()

    for favorite in favorite_users:
        user = favorite.user

        notification = await Notification.objects.acreate(
            kind=NotificationKindEnum.INFO,
        )

        for language in languages:
            await sync_to_async(notification.set_current_language)(language)
            if language == "en":
                await sync_to_async(setattr)(
                    notification, "title", "Price Drop!"
                )
                name = await get_instance_name()
                await sync_to_async(setattr)(
                    notification,
                    "message",
                    f"The price of <a href='{product_url}'>{name}</a> has dropped"
                    f" from {old_price} to {new_price}. Check it out now!",
                )
            elif language == "el":
                await sync_to_async(setattr)(
                    notification, "title", "Μείωση Τιμής!"
                )
                name = await get_instance_name()
                await sync_to_async(setattr)(
                    notification,
                    "message",
                    f"Η τιμή του <a href='{product_url}'>{name}</a> μειώθηκε"  # noqa: RUF001
                    f" από {old_price} σε {new_price}. Δείτε το τώρα!",
                )
            await notification.asave()

        await NotificationUser.objects.acreate(
            user=user, notification=notification
        )



# Generated by Django 5.1 on 2024-08-24 22:51
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("contact", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="contact",
            name="email",
            field=models.EmailField(max_length=254, verbose_name="Email"),
        ),
        migrations.AddIndex(
            model_name="contact",
            index=models.Index(fields=["email"], name="contact_con_email_2ebc86_idx"),
        ),
    ]



from __future__ import annotations

from django.db import models


class UserAddressManager(models.Manager):
    def get_user_addresses(self, user):
        return self.filter(user=user).order_by("-is_main", "title")

    def get_main_address(self, user):
        return self.filter(user=user, is_main=True).first()



from typing import TYPE_CHECKING, Literal, NamedTuple

from meili._client import client

if TYPE_CHECKING:
    from .models import IndexMixin


class Radius(NamedTuple):
    lat: float | str
    lng: float | str
    radius: int


class BoundingBox(NamedTuple):
    top_right: tuple[float | str, float | str]
    bottom_left: tuple[float | str, float | str]


class Point(NamedTuple):
    lat: float | str
    lng: float | str


class IndexQuerySet:
    def __init__(self, model: type["IndexMixin"]):
        self.model = model
        self.index = client.get_index(model._meilisearch["index_name"])
        self.__offset = 0
        self.__limit = 20
        self.__filters: list[str] = []
        self.__facets: list[str] = []
        self.__attributes_to_retrieve: list[str] = ["*"]
        self.__attributes_to_crop: list[str] = []
        self.__crop_length: int = 10
        self.__crop_marker: str = "..."
        self.__attributes_to_highlight: list[str] = ["*"]
        self.__highlight_pre_tag: str = "<mark>"
        self.__highlight_post_tag: str = "</mark>"
        self.__show_matches_position: bool = False
        self.__sort: list[str] = []
        self.__matching_strategy: Literal["last", "all"] = "last"
        self.__show_ranking_score: bool = False
        self.__attributes_to_search_on: list[str] = ["*"]
        self.__locales: list[str] = []

    def __repr__(self):
        return f"<IndexQuerySet for {self.model.__name__}>"

    def __str__(self):
        return f"IndexQuerySet for {self.model.__name__}"

    def __getitem__(self, index):
        if isinstance(index, slice):
            self.__offset = index.start
            self.__limit = index.stop
            return self
        else:
            raise TypeError("IndexQuerySet indices must be slices")

    def count(self) -> int:
        return self.index.get_stats().number_of_documents

    def paginate(self, limit: int, offset: int):
        self.__limit = limit
        self.__offset = offset
        return self

    def order_by(self, *fields: str):
        for field in fields:
            geopoint = "_" if "geoPoint" in field else ""
            if field.startswith("-"):
                self.__sort.append(f"{geopoint}{field[1:]}:desc")
            else:
                self.__sort.append(f"{geopoint}{field}:asc")
        return self

    def filter(self, *geo_filters, **filters):
        self._apply_geo_filters(*geo_filters)
        self._apply_regular_filters(**filters)
        return self

    def _apply_geo_filters(self, *geo_filters):
        for geo_filter in geo_filters:
            if not self.model._meilisearch["supports_geo"]:
                raise TypeError(
                    f"Model {self.model.__name__} does not support geo filters"
                )
            if not isinstance(geo_filter, Radius | BoundingBox):
                raise TypeError(
                    f"Unnamed Argument must be of type Radius or BoundingBox, not {type(geo_filter)}"
                )
            if isinstance(geo_filter, Radius):
                self.__filters.append(
                    f"_geoRadius({geo_filter.lat}, {geo_filter.lng}, {geo_filter.radius})"
                )
            elif isinstance(geo_filter, BoundingBox):
                self.__filters.append(
                    f"_geoBoundingBox([{geo_filter.top_right[0]}, {geo_filter.top_right[1]}], [{geo_filter.bottom_left[0]}, {geo_filter.bottom_left[1]}])"
                )
        return self

    def _apply_regular_filters(self, **filters):
        for full_lookup, value in filters.items():
            if "__" not in full_lookup or "__exact" in full_lookup:
                if (
                    value == ""
                    or (isinstance(value, list) and len(value) == 0)
                    or value == {}
                ):
                    self.__filters.append(
                        f"{full_lookup.split('__')[0]} IS EMPTY"
                    )
                elif value is None:
                    self.__filters.append(
                        f"{full_lookup.split('__')[0]} IS NULL"
                    )
                else:
                    self.__filters.append(
                        f"{full_lookup.split('__')[0]} = '{value}'"
                        if isinstance(value, str)
                        else f"{full_lookup.split('__')[0]} = {value}"
                    )
            elif "__gte" in full_lookup:
                if not isinstance(value, int | float):
                    raise TypeError(
                        f"Cannot compare {type(value)} with int or float"
                    )
                self.__filters.append(
                    f"{full_lookup.split('__')[0]} >= {value}"
                )
            elif "__gt" in full_lookup:
                if not isinstance(value, int | float):
                    raise TypeError(
                        f"Cannot compare {type(value)} with int or float"
                    )
                self.__filters.append(f"{full_lookup.split('__')[0]} > {value}")
            elif "__lte" in full_lookup:
                if not isinstance(value, int | float):
                    raise TypeError(
                        f"Cannot compare {type(value)} with int or float"
                    )
                self.__filters.append(
                    f"{full_lookup.split('__')[0]} <= {value}"
                )
            elif "__lt" in full_lookup:
                if not isinstance(value, int | float):
                    raise TypeError(
                        f"Cannot compare {type(value)} with int or float"
                    )
                self.__filters.append(f"{full_lookup.split('__')[0]} < {value}")
            elif "__in" in full_lookup:
                if not isinstance(value, list):
                    raise TypeError(f"Cannot compare {type(value)} with list")
                self.__filters.append(
                    f"{full_lookup.split('__')[0]} IN {value}"
                )
            elif "__range" in full_lookup:
                if not isinstance(value, range | list | tuple):
                    raise TypeError(
                        f"Cannot compare {type(value)} with range, list or tuple"
                    )
                self.__filters.append(
                    f"{full_lookup.split('__')[0]} {value[0]} TO {value[1]}"
                    if not isinstance(value, range)
                    else f"{full_lookup.split('__')[0]} {value.start} TO {value.stop}"
                )
            elif "__exists" in full_lookup:
                if not isinstance(value, bool):
                    raise TypeError(f"Cannot compare {type(value)} with bool")
                self.__filters.append(
                    f"{full_lookup.split('__')[0]} {'NOT ' if not value else ''}EXISTS"
                )
            elif "__isnull" in full_lookup:
                if not isinstance(value, bool):
                    raise TypeError(f"Cannot compare {type(value)} with bool")
                self.__filters.append(
                    f"{full_lookup.split('__')[0]} {'NOT ' if not value else ''}IS NULL"
                )

    def matching_strategy(self, strategy: Literal["last", "all"]):
        self.__matching_strategy = strategy
        return self

    def attributes_to_search_on(self, *attributes):
        self.__attributes_to_search_on.append(*attributes)
        return self

    def search(self, q: str = ""):
        results = self.index.search(
            q,
            {
                "filter": self.__filters,
                "facets": self.__facets,
                "offset": self.__offset,
                "limit": self.__limit,
                "attributesToRetrieve": self.__attributes_to_retrieve,
                "attributesToCrop": self.__attributes_to_crop,
                "cropLength": self.__crop_length,
                "cropMarker": self.__crop_marker,
                "attributesToHighlight": self.__attributes_to_highlight,
                "highlightPreTag": self.__highlight_pre_tag,
                "highlightPostTag": self.__highlight_post_tag,
                "showMatchesPosition": self.__show_matches_position,
                "sort": self.__sort,
                "matchingStrategy": self.__matching_strategy,
                "showRankingScore": self.__show_ranking_score,
                "attributesToSearchOn": self.__attributes_to_search_on,
                "locales": self.__locales,
            },
        )

        id_field = getattr(self.model.MeiliMeta, "primary_key", "id")
        hits = results.get("hits", [])
        filtered_objects = self.model.objects.filter(
            pk__in=[hit[id_field] for hit in hits]
        )

        enriched_results = []
        for obj in filtered_objects:
            hit = next(hit for hit in hits if str(hit["id"]) == str(obj.pk))
            enriched_results.append(
                {
                    "object": obj,
                    "_formatted": hit.get("_formatted", {}),
                    "_matchesPosition": hit.get("_matchesPosition", {}),
                    "_rankingScore": hit.get("_rankingScore", None),
                }
            )

        return {
            "results": enriched_results,
            "estimated_total_hits": results["estimatedTotalHits"],
            "offset": self.__offset,
            "limit": self.__limit,
        }



from django.contrib.auth import get_user_model
from django.urls import reverse
from datetime import timedelta
from django.utils import timezone
from rest_framework.test import APITestCase

from notification.factories.notification import NotificationFactory
from notification.factories.user import NotificationUserFactory
from notification.models.user import NotificationUser
from notification.enum import (
    NotificationKindEnum,
    NotificationCategoryEnum,
    NotificationPriorityEnum,
)
from user.factories.account import UserAccountFactory

User = get_user_model()


class NotificationUserFilterTest(APITestCase):
    def setUp(self):
        NotificationUser.objects.all().delete()

        self.user1 = UserAccountFactory(
            email="user1@example.com",
            first_name="John",
            last_name="Doe",
            is_active=True,
            is_staff=False,
        )
        self.user2 = UserAccountFactory(
            email="user2@example.com",
            first_name="Jane",
            last_name="Smith",
            is_active=True,
            is_staff=True,
        )
        self.inactive_user = UserAccountFactory(
            email="inactive@example.com",
            first_name="Inactive",
            last_name="User",
            is_active=False,
            is_staff=False,
        )

        self.now = timezone.now()

        self.high_priority_notification = NotificationFactory(
            kind=NotificationKindEnum.ERROR,
            category=NotificationCategoryEnum.ORDER,
            priority=NotificationPriorityEnum.HIGH,
            notification_type="order_failed",
            link="https://example.com/order/123",
            expiry_date=self.now + timedelta(days=30),
        )
        self.high_priority_notification.created_at = self.now - timedelta(
            hours=1
        )
        self.high_priority_notification.save()

        self.normal_notification = NotificationFactory(
            kind=NotificationKindEnum.INFO,
            category=NotificationCategoryEnum.SYSTEM,
            priority=NotificationPriorityEnum.NORMAL,
            notification_type="system_update",
            link="https://example.com/system",
            expiry_date=self.now + timedelta(days=7),
        )
        self.normal_notification.created_at = self.now - timedelta(days=2)
        self.normal_notification.save()

        self.expired_notification = NotificationFactory(
            kind=NotificationKindEnum.WARNING,
            category=NotificationCategoryEnum.PAYMENT,
            priority=NotificationPriorityEnum.NORMAL,
            notification_type="payment_reminder",
            link="https://example.com/payment",
            expiry_date=self.now - timedelta(days=1),
        )
        self.expired_notification.created_at = self.now - timedelta(days=10)
        self.expired_notification.save()

        self.old_notification = NotificationFactory(
            kind=NotificationKindEnum.SUCCESS,
            category=NotificationCategoryEnum.ACCOUNT,
            priority=NotificationPriorityEnum.LOW,
            notification_type="account_verified",
            link="",
            expiry_date=self.now + timedelta(days=60),
        )
        self.old_notification.created_at = self.now - timedelta(days=30)
        self.old_notification.save()

        self.seen_notification_user = NotificationUserFactory(
            user=self.user1,
            notification=self.high_priority_notification,
            seen=True,
            seen_at=self.now - timedelta(minutes=30),
        )

        self.unseen_notification_user = NotificationUserFactory(
            user=self.user1,
            notification=self.normal_notification,
            seen=False,
            seen_at=None,
        )

        self.recent_seen_notification_user = NotificationUserFactory(
            user=self.user2,
            notification=self.normal_notification,
            seen=True,
            seen_at=self.now - timedelta(hours=2),
        )

        self.old_unseen_notification_user = NotificationUserFactory(
            user=self.user2,
            notification=self.old_notification,
            seen=False,
            seen_at=None,
        )

        self.expired_notification_user = NotificationUserFactory(
            user=self.inactive_user,
            notification=self.expired_notification,
            seen=False,
            seen_at=None,
        )

    def test_basic_filters(self):
        url = reverse("notification-user-list")

        response = self.client.get(url, {"user": self.user1.id})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.seen_notification_user.id, result_ids)
        self.assertIn(self.unseen_notification_user.id, result_ids)

        response = self.client.get(
            url, {"notification": self.normal_notification.id}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        response = self.client.get(url, {"seen": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.seen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        response = self.client.get(url, {"seen": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 3)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.old_unseen_notification_user.id, result_ids)
        self.assertIn(self.expired_notification_user.id, result_ids)

    def test_user_relationship_filters(self):
        url = reverse("notification-user-list")

        response = self.client.get(url, {"user__email": "user1"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.seen_notification_user.id, result_ids)
        self.assertIn(self.unseen_notification_user.id, result_ids)

        response = self.client.get(url, {"user__first_name": "jane"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)
        self.assertIn(self.old_unseen_notification_user.id, result_ids)

        response = self.client.get(url, {"user__is_active": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.expired_notification_user.id, result_ids)

        response = self.client.get(url, {"user__is_staff": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)
        self.assertIn(self.old_unseen_notification_user.id, result_ids)

    def test_seen_timestamp_filters(self):
        url = reverse("notification-user-list")

        seen_after_date = self.now - timedelta(hours=1)
        response = self.client.get(
            url, {"seen_after": seen_after_date.isoformat()}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.seen_notification_user.id, result_ids)

        seen_before_date = self.now - timedelta(hours=1)
        response = self.client.get(
            url, {"seen_before": seen_before_date.isoformat()}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        response = self.client.get(url, {"has_seen_at": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.seen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        response = self.client.get(url, {"has_seen_at": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 3)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.old_unseen_notification_user.id, result_ids)
        self.assertIn(self.expired_notification_user.id, result_ids)

    def test_notification_relationship_filters(self):
        url = reverse("notification-user-list")

        response = self.client.get(
            url, {"notification__kind": NotificationKindEnum.ERROR}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.seen_notification_user.id, result_ids)

        response = self.client.get(
            url, {"notification__category": NotificationCategoryEnum.SYSTEM}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        response = self.client.get(
            url, {"notification__priority": NotificationPriorityEnum.HIGH}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.seen_notification_user.id, result_ids)

        response = self.client.get(url, {"notification__type": "order"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.seen_notification_user.id, result_ids)

        response = self.client.get(
            url, {"notification__link": "example.com/system"}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

    def test_notification_expiry_filters(self):
        url = reverse("notification-user-list")

        response = self.client.get(url, {"notification__is_expired": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.expired_notification_user.id, result_ids)

        response = self.client.get(url, {"notification__is_expired": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 4)
        self.assertNotIn(self.expired_notification_user.id, result_ids)

        expires_after_date = self.now + timedelta(days=10)
        response = self.client.get(
            url, {"notification__expires_after": expires_after_date.isoformat()}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.seen_notification_user.id, result_ids)
        self.assertIn(self.old_unseen_notification_user.id, result_ids)

        expires_before_date = self.now + timedelta(days=10)
        response = self.client.get(
            url,
            {"notification__expires_before": expires_before_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 3)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)
        self.assertIn(self.expired_notification_user.id, result_ids)

    def test_special_filters(self):
        url = reverse("notification-user-list")

        response = self.client.get(url, {"unseen_only": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 3)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.old_unseen_notification_user.id, result_ids)
        self.assertIn(self.expired_notification_user.id, result_ids)

        response = self.client.get(url, {"seen_only": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.seen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        response = self.client.get(url, {"recent_notifications": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 3)
        self.assertIn(self.seen_notification_user.id, result_ids)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        response = self.client.get(url, {"high_priority": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.seen_notification_user.id, result_ids)

    def test_bulk_filters(self):
        url = reverse("notification-user-list")

        user_ids = f"{self.user1.id},{self.user2.id}"
        response = self.client.get(url, {"user_ids": user_ids})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 4)
        self.assertNotIn(self.expired_notification_user.id, result_ids)

        notification_ids = f"{self.high_priority_notification.id},{self.normal_notification.id}"
        response = self.client.get(url, {"notification_ids": notification_ids})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 3)
        self.assertIn(self.seen_notification_user.id, result_ids)
        self.assertIn(self.unseen_notification_user.id, result_ids)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        response = self.client.get(url, {"user_ids": "invalid,ids"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data["results"]), 0)

    def test_uuid_filter(self):
        url = reverse("notification-user-list")

        response = self.client.get(
            url, {"uuid": str(self.seen_notification_user.uuid)}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(
            response.data["results"][0]["id"], self.seen_notification_user.id
        )

    def test_timestamp_filters(self):
        url = reverse("notification-user-list")

        created_after_date = self.now - timedelta(minutes=5)
        response = self.client.get(
            url, {"created_after": created_after_date.isoformat()}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 5)

        created_before_date = self.now + timedelta(minutes=5)
        response = self.client.get(
            url, {"created_before": created_before_date.isoformat()}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 5)

    def test_camel_case_filters(self):
        url = reverse("notification-user-list")

        response = self.client.get(
            url,
            {
                "user__email": "user1",
                "notification__kind": NotificationKindEnum.ERROR,
                "unseen_only": "true",
            },
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 0)

        response = self.client.get(
            url,
            {
                "user__email": "user1",
                "notification__category": NotificationCategoryEnum.SYSTEM,
                "unseen_only": "true",
            },
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.unseen_notification_user.id, result_ids)

    def test_complex_filter_combinations(self):
        url = reverse("notification-user-list")

        response = self.client.get(
            url,
            {
                "user__is_active": "true",
                "seen": "false",
                "notification__priority": NotificationPriorityEnum.HIGH,
                "ordering": "-created_at",
            },
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 0)

        response = self.client.get(
            url,
            {
                "user__is_staff": "true",
                "recent_notifications": "true",
                "ordering": "-notification__created_at",
            },
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.recent_seen_notification_user.id, result_ids)

        created_after_date = self.now - timedelta(days=5)
        expires_after_date = self.now + timedelta(days=5)
        response = self.client.get(
            url,
            {
                "created_after": created_after_date.isoformat(),
                "notification__expires_after": expires_after_date.isoformat(),
                "seen": "true",
                "ordering": "-seen_at",
            },
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 1)

    def test_filter_with_ordering(self):
        url = reverse("notification-user-list")

        response = self.client.get(
            url, {"seen": "true", "ordering": "-seen_at"}
        )
        self.assertEqual(response.status_code, 200)
        results = response.data["results"]
        self.assertEqual(len(results), 2)

        self.assertEqual(results[0]["id"], self.seen_notification_user.id)
        self.assertEqual(
            results[1]["id"], self.recent_seen_notification_user.id
        )

        response = self.client.get(
            url,
            {"user": self.user1.id, "ordering": "-notification__created_at"},
        )
        self.assertEqual(response.status_code, 200)
        results = response.data["results"]
        self.assertEqual(len(results), 2)

        self.assertEqual(results[0]["id"], self.seen_notification_user.id)
        self.assertEqual(results[1]["id"], self.unseen_notification_user.id)

    def tearDown(self):
        NotificationUser.objects.all().delete()



from unittest.mock import MagicMock, patch

import pytest

from meili._client import Client
from meili._client import client as global_client
from meili._settings import _MeiliSettings
from meili.dataclasses import MeiliIndexSettings


class TestMeiliClient:
    def setup_method(self):
        self.settings = _MeiliSettings(
            host="localhost",
            port=7700,
            https=False,
            master_key="test_key",
            timeout=10,
            sync=False,
            client_agents=("test-agent",),
            debug=False,
            offline=False,
            batch_size=1000,
        )

    @patch("meili._client._Client")
    def test_client_initialization(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_client_class.return_value = mock_client_instance

        client = Client(self.settings)

        assert client.client == mock_client_instance
        assert client.is_sync == self.settings.sync
        assert client.tasks == []

        mock_client_class.assert_called_once_with(
            "http://localhost:7700",
            "test_key",
            timeout=10,
            client_agents=("test-agent",),
        )

    @patch("meili._client._Client")
    def test_client_initialization_https(self, mock_client_class):
        https_settings = _MeiliSettings(
            host="example.com",
            port=443,
            https=True,
            master_key="https_key",
            timeout=30,
            sync=True,
            client_agents=None,
            debug=False,
            offline=False,
            batch_size=1000,
        )

        client = Client(https_settings)

        mock_client_class.assert_called_once_with(
            "https://example.com:443",
            "https_key",
            timeout=30,
            client_agents=None,
        )
        assert client.is_sync is True

    def test_flush_tasks(self):
        client = Client(self.settings)
        client.tasks = ["task1", "task2", "task3"]

        client.flush_tasks()

        assert client.tasks == []

    @patch("meili._client._Client")
    def test_with_settings_basic(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_index = MagicMock()
        mock_update_settings = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.index.return_value = mock_index
        mock_index.update_settings.return_value = mock_update_settings

        client = Client(self.settings)
        index_settings = MeiliIndexSettings()

        result = client.with_settings("test_index", index_settings)

        assert result == client

        mock_index.update_settings.assert_called_once()
        call_args = mock_index.update_settings.call_args[0][0]

        expected_settings = {
            "displayedAttributes": ["*"],
            "searchableAttributes": ["*"],
            "filterableAttributes": [],
            "sortableAttributes": [],
            "rankingRules": [
                "words",
                "typo",
                "proximity",
                "attribute",
                "sort",
                "exactness",
            ],
            "stopWords": [],
            "synonyms": {},
            "distinctAttribute": None,
            "typoTolerance": {
                "enabled": True,
                "minWordSizeForTypos": {"oneTypo": 5, "twoTypos": 9},
                "disableOnWords": [],
                "disableOnAttributes": [],
            },
            "faceting": {"maxValuesPerFacet": 100},
            "pagination": {"maxTotalHits": 1000},
        }

        assert call_args == expected_settings

    @patch("meili._client._Client")
    def test_with_settings_custom(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_index = MagicMock()
        mock_update_settings = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.index.return_value = mock_index
        mock_index.update_settings.return_value = mock_update_settings

        client = Client(self.settings)

        index_settings = MeiliIndexSettings(
            displayed_fields=["id", "title"],
            searchable_fields=["title", "content"],
            filterable_fields=["category", "status"],
            sortable_fields=["created_at", "updated_at"],
            ranking_rules=["words", "typo", "custom"],
            stop_words=["the", "and", "or"],
            synonyms={"car": ["automobile", "vehicle"]},
            distinct_attribute="id",
            typo_tolerance={"enabled": False},
            faceting={"maxValuesPerFacet": 50},
            pagination={"maxTotalHits": 500},
        )

        client.with_settings("custom_index", index_settings)

        call_args = mock_index.update_settings.call_args[0][0]

        assert call_args["displayedAttributes"] == ["id", "title"]
        assert call_args["searchableAttributes"] == ["title", "content"]
        assert call_args["filterableAttributes"] == ["category", "status"]
        assert call_args["sortableAttributes"] == ["created_at", "updated_at"]
        assert call_args["rankingRules"] == ["words", "typo", "custom"]
        assert call_args["stopWords"] == ["the", "and", "or"]
        assert call_args["synonyms"] == {"car": ["automobile", "vehicle"]}
        assert call_args["distinctAttribute"] == "id"
        assert call_args["typoTolerance"] == {"enabled": False}
        assert call_args["faceting"] == {"maxValuesPerFacet": 50}
        assert call_args["pagination"] == {"maxTotalHits": 500}

    @patch("meili._client._Client")
    def test_create_index_new(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_create_index = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.create_index.return_value = mock_create_index
        mock_client_instance.get_indexes.return_value = {"results": []}

        client = Client(self.settings)

        result = client.create_index("new_index", "id")

        assert result == client

        mock_client_instance.create_index.assert_called_once_with(
            "new_index", {"primaryKey": "id"}
        )
        assert len(client.tasks) == 1

    @patch("meili._client._Client")
    def test_create_index_existing(self, mock_client_class):
        mock_client_instance = MagicMock()
        existing_index = MagicMock()
        existing_index.uid = "existing_index"

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.get_indexes.return_value = {
            "results": [existing_index]
        }

        client = Client(self.settings)

        result = client.create_index("existing_index", "id")

        assert result == client

        mock_client_instance.create_index.assert_not_called()
        assert len(client.tasks) == 0

    @patch("meili._client._Client")
    def test_get_index(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_index = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.index.return_value = mock_index

        client = Client(self.settings)

        result = client.get_index("test_index")

        assert result == mock_index
        mock_client_instance.index.assert_called_once_with("test_index")

    @patch("meili._client._Client")
    def test_wait_for_task(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_task = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.wait_for_task.return_value = mock_task

        client = Client(self.settings)

        result = client.wait_for_task(123)

        assert result == mock_task
        mock_client_instance.wait_for_task.assert_called_once_with(123)

    @patch("meili._client._Client")
    def test_get_indexes(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_indexes = [{"uid": "index1"}, {"uid": "index2"}]

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.get_indexes.return_value = {
            "results": mock_indexes
        }

        client = Client(self.settings)

        result = client.get_indexes()

        assert result == mock_indexes
        mock_client_instance.get_indexes.assert_called_once()

    @patch("meili._client._Client")
    def test_update_display_attributes(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_index = MagicMock()
        mock_update = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.index.return_value = mock_index
        mock_index.update_displayed_attributes.return_value = mock_update

        client = Client(self.settings)

        result = client.update_display("test_index", ["id", "title"])

        assert result == client
        mock_index.update_displayed_attributes.assert_called_once_with(
            ["id", "title"]
        )

    @patch("meili._client._Client")
    def test_update_display_attributes_none(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_client_class.return_value = mock_client_instance

        client = Client(self.settings)

        result = client.update_display("test_index", None)

        assert result == client

    @patch("meili._client._Client")
    def test_update_searchable_attributes(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_index = MagicMock()
        mock_update = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.index.return_value = mock_index
        mock_index.update_searchable_attributes.return_value = mock_update

        client = Client(self.settings)

        result = client.update_searchable("test_index", ["title", "content"])

        assert result == client
        mock_index.update_searchable_attributes.assert_called_once_with(
            ["title", "content"]
        )

    @patch("meili._client._Client")
    def test_update_filterable_attributes(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_index = MagicMock()
        mock_update = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.index.return_value = mock_index
        mock_index.update_filterable_attributes.return_value = mock_update

        client = Client(self.settings)

        result = client.update_filterable("test_index", ["category", "status"])

        assert result == client
        mock_index.update_filterable_attributes.assert_called_once_with(
            ["category", "status"]
        )

    @patch("meili._client._Client")
    def test_update_sortable_attributes(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_index = MagicMock()
        mock_update = MagicMock()

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.index.return_value = mock_index
        mock_index.update_sortable_attributes.return_value = mock_update

        client = Client(self.settings)

        result = client.update_sortable(
            "test_index", ["created_at", "updated_at"]
        )

        assert result == client
        mock_index.update_sortable_attributes.assert_called_once_with(
            ["created_at", "updated_at"]
        )

    @patch("meili._client._Client")
    def test_handle_sync_task_uid(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_task = MagicMock()
        mock_waited_task = MagicMock()

        mock_task.task_uid = 123
        mock_waited_task.status = "succeeded"

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.wait_for_task.return_value = mock_waited_task

        sync_settings = _MeiliSettings(
            host="localhost",
            port=7700,
            https=False,
            master_key="test_key",
            sync=True,
            timeout=None,
            client_agents=None,
            debug=False,
            offline=False,
            batch_size=1000,
        )

        client = Client(sync_settings)

        result = client._handle_sync(mock_task)

        assert result == mock_waited_task
        mock_client_instance.wait_for_task.assert_called_once_with(123)

    @patch("meili._client._Client")
    def test_handle_sync_uid(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_task = MagicMock()
        mock_waited_task = MagicMock()

        del mock_task.task_uid
        mock_task.uid = 456
        mock_waited_task.status = "succeeded"

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.wait_for_task.return_value = mock_waited_task

        sync_settings = _MeiliSettings(
            host="localhost",
            port=7700,
            https=False,
            master_key="test_key",
            sync=True,
            timeout=None,
            client_agents=None,
            debug=False,
            offline=False,
            batch_size=1000,
        )

        client = Client(sync_settings)

        with patch("builtins.hasattr") as mock_hasattr:

            def hasattr_side_effect(obj, attr):
                if attr == "task_uid":
                    return False
                elif attr == "uid":
                    return True
                return False

            mock_hasattr.side_effect = hasattr_side_effect

            result = client._handle_sync(mock_task)

        assert result == mock_waited_task
        mock_client_instance.wait_for_task.assert_called_once_with(456)

    @patch("meili._client._Client")
    def test_handle_sync_no_uid(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_task = MagicMock()

        mock_client_class.return_value = mock_client_instance

        sync_settings = _MeiliSettings(
            host="localhost",
            port=7700,
            https=False,
            master_key="test_key",
            sync=True,
            timeout=None,
            client_agents=None,
            debug=False,
            offline=False,
            batch_size=1000,
        )

        client = Client(sync_settings)

        with patch("builtins.hasattr") as mock_hasattr:
            mock_hasattr.return_value = False

            with pytest.raises(
                AttributeError, match="Task object has no uid attribute"
            ):
                client._handle_sync(mock_task)

    @patch("meili._client._Client")
    def test_handle_sync_failed_task(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_task = MagicMock()
        mock_waited_task = MagicMock()

        mock_task.task_uid = 789
        mock_waited_task.status = "failed"
        mock_waited_task.error = "Test error message"

        mock_client_class.return_value = mock_client_instance
        mock_client_instance.wait_for_task.return_value = mock_waited_task

        sync_settings = _MeiliSettings(
            host="localhost",
            port=7700,
            https=False,
            master_key="test_key",
            sync=True,
            timeout=None,
            client_agents=None,
            debug=False,
            offline=False,
            batch_size=1000,
        )

        client = Client(sync_settings)

        with pytest.raises(Exception, match="Test error message"):
            client._handle_sync(mock_task)

    @patch("meili._client._Client")
    def test_handle_sync_async_mode(self, mock_client_class):
        mock_client_instance = MagicMock()
        mock_task = MagicMock()

        mock_client_class.return_value = mock_client_instance

        client = Client(self.settings)

        result = client._handle_sync(mock_task)

        assert result == mock_task
        mock_client_instance.wait_for_task.assert_not_called()

    @patch("meili._settings._MeiliSettings.from_settings")
    @patch("meili._client._Client")
    def test_global_client_instance(
        self, mock_client_class, mock_from_settings
    ):
        assert isinstance(global_client, Client)



from __future__ import annotations

from django.utils.translation import gettext_lazy as _
from drf_spectacular.openapi import OpenApiTypes
from drf_spectacular.utils import (
    OpenApiParameter,
    extend_schema,
    extend_schema_view,
)
from rest_framework import status
from rest_framework.permissions import IsAdminUser

from rest_framework.response import Response

from cart.filters.cart import CartFilter
from cart.models import Cart
from cart.serializers.cart import (
    CartDetailSerializer,
    CartSerializer,
    CartWriteSerializer,
)
from cart.services import CartService
from core.api.serializers import ErrorResponseSerializer
from core.api.views import BaseModelViewSet

from core.utils.serializers import (
    create_schema_view_config,
    RequestSerializersConfig,
    ResponseSerializersConfig,
)

GUEST_CART_HEADERS = [
    OpenApiParameter(
        name="X-Cart-Id",
        type=OpenApiTypes.INT,
        location=OpenApiParameter.HEADER,
        description="Cart ID for guest users. Used to identify and maintain guest cart sessions.",
        required=False,
    ),
    OpenApiParameter(
        name="X-Session-Key",
        type=OpenApiTypes.STR,
        location=OpenApiParameter.HEADER,
        description="Session key for guest users. Used to validate cart ownership for anonymous sessions.",
        required=False,
    ),
]

req_serializers: RequestSerializersConfig = {
    "update": CartWriteSerializer,
    "partial_update": CartWriteSerializer,
}

res_serializers: ResponseSerializersConfig = {
    "list": CartSerializer,
    "retrieve": CartDetailSerializer,
    "update": CartDetailSerializer,
    "partial_update": CartDetailSerializer,
}

cart_schema_config = create_schema_view_config(
    model_class=Cart,
    request_serializers=req_serializers,
    response_serializers=res_serializers,
    error_serializer=ErrorResponseSerializer,
    display_config={
        "tag": "Cart",
        "display_name": _("cart"),
        "display_name_plural": _("carts"),
    },
)

cart_schema_config.update(
    {
        "list": extend_schema(
            operation_id="listCart",
            summary=_("Get cart"),
            description=_(
                "Get a cart. For guest users, include X-Cart-Id and X-Session-Key headers to maintain cart session."
            ),
            tags=["Cart"],
            parameters=GUEST_CART_HEADERS,
            responses={
                200: CartSerializer(many=True),
                404: ErrorResponseSerializer,
            },
        ),
        "retrieve": extend_schema(
            operation_id="retrieveCart",
            summary=_("Get cart"),
            description=_(
                "Get a cart. For guest users, include X-Cart-Id and X-Session-Key headers to maintain cart session."
            ),
            tags=["Cart"],
            parameters=GUEST_CART_HEADERS,
            responses={
                200: CartDetailSerializer,
                404: ErrorResponseSerializer,
            },
        ),
        "update": extend_schema(
            operation_id="updateCart",
            summary=_("Update cart"),
            description=_(
                "Update a cart. For guest users, include X-Cart-Id and X-Session-Key headers to maintain cart session."
            ),
            tags=["Cart"],
            parameters=GUEST_CART_HEADERS,
            request=CartWriteSerializer,
            responses={
                200: CartDetailSerializer,
                404: ErrorResponseSerializer,
            },
        ),
        "partial_update": extend_schema(
            operation_id="partialUpdateCart",
            summary=_("Update cart"),
            description=_(
                "Update a cart. For guest users, include X-Cart-Id and X-Session-Key headers to maintain cart session."
            ),
            tags=["Cart"],
            parameters=GUEST_CART_HEADERS,
            request=CartWriteSerializer,
            responses={
                200: CartDetailSerializer,
                404: ErrorResponseSerializer,
            },
        ),
        "destroy": extend_schema(
            operation_id="destroyCart",
            summary=_("Delete cart"),
            description=_(
                "Delete a cart. For guest users, include X-Cart-Id and X-Session-Key headers to maintain cart session."
            ),
            tags=["Cart"],
            parameters=GUEST_CART_HEADERS,
            responses={
                204: None,
                401: ErrorResponseSerializer,
                404: ErrorResponseSerializer,
            },
        ),
        "create": extend_schema(
            operation_id="createCart",
            summary=_("Create cart"),
            description=_("Cart creation is not allowed via API."),
            tags=["Cart"],
            responses={
                405: ErrorResponseSerializer,
            },
        ),
    }
)


@extend_schema_view(**cart_schema_config)
class CartViewSet(BaseModelViewSet):
    cart_service: CartService
    queryset = Cart.objects.all()
    response_serializers = res_serializers
    request_serializers = req_serializers
    filterset_class = CartFilter
    ordering_fields = [
        "id",
        "user",
        "created_at",
        "updated_at",
        "last_activity",
    ]
    ordering = ["-last_activity", "-created_at"]
    search_fields = ["user__email", "user__first_name", "user__last_name"]

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        self.cart_service = CartService(request=request)

    def get_permissions(self):
        if self.action in [
            "list",
        ]:
            self.permission_classes = [IsAdminUser]
        return super().get_permissions()

    def get_queryset(self):
        user = self.request.user

        if user.is_staff:
            return Cart.objects.all()
        elif user.is_authenticated:
            return Cart.objects.filter(user=user)
        elif self.cart_service.cart:
            return Cart.objects.filter(id=self.cart_service.cart.id)

        return Cart.objects.none()

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["cart"] = self.cart_service.cart
        return context

    def create(self, request, *args, **kwargs):
        return Response(status=status.HTTP_405_METHOD_NOT_ALLOWED)

    def retrieve(self, request, *args, **kwargs):
        cart = self.cart_service.get_or_create_cart()
        if not cart:
            return Response(status=status.HTTP_404_NOT_FOUND)
        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(cart)
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    def update(self, request, *args, **kwargs):
        cart = self.cart_service.get_or_create_cart()
        if not cart:
            return Response(status=status.HTTP_404_NOT_FOUND)
        request_serializer_class = self.get_request_serializer()
        request_serializer = request_serializer_class(
            cart, data=request.data, partial=kwargs.pop("partial", False)
        )
        request_serializer.is_valid(raise_exception=True)
        self.perform_update(request_serializer)

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(
            cart, context=self.get_serializer_context()
        )
        return Response(response_serializer.data)

    def partial_update(self, request, *args, **kwargs):
        kwargs["partial"] = True
        return self.update(request, *args, **kwargs)

    def destroy(self, request, *args, **kwargs):
        cart = self.cart_service.get_or_create_cart()
        if cart:
            cart.delete()
            return Response(status=status.HTTP_204_NO_CONTENT)
        return Response(status=status.HTTP_404_NOT_FOUND)



# Generated by Django 5.1 on 2024-08-24 22:51
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("pay_way", "0005_alter_payway_cost_alter_payway_free_for_order_amount"),
    ]

    operations = [
        migrations.AlterField(
            model_name="payway",
            name="sort_order",
            field=models.IntegerField(
                editable=False, null=True, verbose_name="Sort Order"
            ),
        ),
    ]



from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase

from product.factories.favourite import ProductFavouriteFactory
from product.factories.product import ProductFactory
from product.models.favourite import ProductFavourite
from product.serializers.favourite import (
    ProductFavouriteDetailSerializer,
    ProductFavouriteSerializer,
)
from user.factories.account import UserAccountFactory

User = get_user_model()


class ProductFavouriteViewSetTestCase(APITestCase):
    def setUp(self):
        self.user = UserAccountFactory(num_addresses=0)
        self.product = ProductFactory(num_images=0, num_reviews=0)
        self.product_favourite = ProductFavouriteFactory(
            product=self.product,
            user=self.user,
        )

    def get_product_favourite_detail_url(self, pk):
        return reverse("product-favourite-detail", args=[pk])

    def get_product_favourite_list_url(self):
        return reverse("product-favourite-list")

    def test_list(self):
        url = self.get_product_favourite_list_url()
        response = self.client.get(url)
        favourites = ProductFavourite.objects.all()
        serializer = ProductFavouriteSerializer(favourites, many=True)

        self.assertEqual(response.data["results"], serializer.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_create_valid(self):
        product_2 = ProductFactory(num_images=0, num_reviews=0)

        payload = {
            "product": product_2.id,
            "user": self.user.id,
        }

        self.client.force_authenticate(user=self.user)

        url = self.get_product_favourite_list_url()
        response = self.client.post(url, payload)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_create_invalid(self):
        payload = {
            "product": "invalid_product",
            "user": "invalid_user",
        }

        url = self.get_product_favourite_list_url()
        response = self.client.post(url, payload)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retrieve_valid(self):
        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.get(url)
        favourite = ProductFavourite.objects.get(pk=self.product_favourite.id)
        serializer = ProductFavouriteDetailSerializer(favourite)

        self.assertEqual(response.data, serializer.data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_retrieve_invalid(self):
        invalid_product_favourite_id = 999999
        url = self.get_product_favourite_detail_url(
            invalid_product_favourite_id
        )
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_uses_correct_serializer(self):
        url = self.get_product_favourite_list_url()
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        if response.data.get("results"):
            favourite_data = response.data["results"][0]
            expected_fields = {
                "id",
                "user_id",
                "user_username",
                "product",
                "created_at",
                "uuid",
            }
            self.assertTrue(
                expected_fields.issubset(set(favourite_data.keys()))
            )

            if "product" in favourite_data and favourite_data["product"]:
                product_fields = {"id", "translations", "price"}
                self.assertTrue(
                    product_fields.issubset(
                        set(favourite_data["product"].keys())
                    )
                )

    def test_retrieve_uses_correct_serializer(self):
        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        expected_fields = {
            "id",
            "user",
            "product",
            "created_at",
            "updated_at",
            "uuid",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_create_request_response_serializers(self):
        product_2 = ProductFactory(num_images=0, num_reviews=0)
        payload = {"product": product_2.id}

        self.client.force_authenticate(user=self.user)
        url = self.get_product_favourite_list_url()
        response = self.client.post(url, payload)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        expected_fields = {"product"}
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

        self.assertEqual(response.data["product"], product_2.id)

    def test_update_request_response_serializers(self):
        payload = {"product": self.product.id}

        self.client.force_authenticate(user=self.user)
        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.put(url, payload)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        expected_fields = {"product"}
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_partial_update_request_response_serializers(self):
        payload = {"product": self.product.id}

        self.client.force_authenticate(user=self.user)
        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.patch(url, payload)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        expected_fields = {"product"}
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_create_with_complex_payload(self):
        product_2 = ProductFactory(num_images=0, num_reviews=0)
        payload = {
            "product": product_2.id,
        }

        self.client.force_authenticate(user=self.user)
        url = self.get_product_favourite_list_url()
        response = self.client.post(url, payload)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["product"], product_2.id)

    def test_validation_errors_consistent(self):
        payload = {"product": 99999}

        self.client.force_authenticate(user=self.user)
        url = self.get_product_favourite_list_url()
        response = self.client.post(url, payload)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("product", response.data)

    def test_duplicate_favourite_validation(self):
        payload = {"product": self.product.id}

        self.client.force_authenticate(user=self.user)
        url = self.get_product_favourite_list_url()
        response = self.client.post(url, payload)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("Product is already in favorites", str(response.data))

    def test_product_action(self):
        url = reverse(
            "product-favourite-product", args=[self.product_favourite.id]
        )
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], self.product.id)
        expected_fields = {"id", "slug", "category", "price", "active", "stock"}
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_favourites_by_products_action(self):
        product_2 = ProductFactory(num_images=0, num_reviews=0)
        ProductFavouriteFactory(user=self.user, product=product_2)

        payload = {"product_ids": [self.product.id, product_2.id]}

        self.client.force_authenticate(user=self.user)
        url = reverse("product-favourite-favourites-by-products")
        response = self.client.post(url, payload)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), 2)

    def test_favourites_by_products_validation(self):
        payload = {}

        self.client.force_authenticate(user=self.user)
        url = reverse("product-favourite-favourites-by-products")
        response = self.client.post(url, payload)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_filtering_functionality(self):
        user_2 = UserAccountFactory(num_addresses=0)
        ProductFavouriteFactory(user=user_2, product=self.product)

        url = self.get_product_favourite_list_url()

        response = self.client.get(url, {"user_id": self.user.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        user_ids = [fav["user_id"] for fav in response.data["results"]]
        self.assertTrue(all(uid == self.user.id for uid in user_ids))

        response = self.client.get(url, {"product_id": self.product.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        product_ids = [fav["product"]["id"] for fav in response.data["results"]]
        self.assertTrue(all(pid == self.product.id for pid in product_ids))

    def test_ordering_functionality(self):
        product_2 = ProductFactory(num_images=0, num_reviews=0)
        ProductFavouriteFactory(user=self.user, product=product_2)

        url = self.get_product_favourite_list_url()

        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        if len(response.data["results"]) >= 2:
            first_created = response.data["results"][0]["created_at"]
            second_created = response.data["results"][1]["created_at"]
            self.assertGreaterEqual(first_created, second_created)

    def test_search_functionality(self):
        url = self.get_product_favourite_list_url()

        response = self.client.get(url, {"search": str(self.user.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_consistency_with_manual_serializer_instantiation(self):
        favourite = ProductFavourite.objects.get(pk=self.product_favourite.id)

        manual_serializer = ProductFavouriteDetailSerializer(favourite)

        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.get(url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["id"], manual_serializer.data["id"])
        self.assertEqual(response.data["uuid"], manual_serializer.data["uuid"])

    def test_queryset_optimization(self):
        url = self.get_product_favourite_list_url()

        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        if response.data.get("results"):
            favourite_data = response.data["results"][0]
            self.assertIsNotNone(favourite_data.get("user_id"))
            self.assertIsNotNone(favourite_data.get("product"))

    def test_update_valid(self):
        payload = {
            "product": self.product.id,
            "user": self.user.id,
        }

        self.client.force_authenticate(user=self.user)

        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.put(url, payload)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_update_invalid(self):
        payload = {
            "product": "invalid_product",
            "user": "invalid_user",
        }

        self.client.force_authenticate(user=self.user)

        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.put(url, payload)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_partial_update_valid(self):
        payload = {
            "product": self.product.id,
        }

        self.client.force_authenticate(user=self.user)

        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.patch(url, payload)

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_partial_update_invalid(self):
        payload = {
            "product": "invalid_product",
        }

        self.client.force_authenticate(user=self.user)

        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.patch(url, payload)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_destroy_valid(self):
        self.client.force_authenticate(user=self.user)

        url = self.get_product_favourite_detail_url(self.product_favourite.id)
        response = self.client.delete(url)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(
            ProductFavourite.objects.filter(
                pk=self.product_favourite.id
            ).exists()
        )

    def test_destroy_invalid(self):
        invalid_product_favourite_id = 999999
        url = self.get_product_favourite_detail_url(
            invalid_product_favourite_id
        )
        response = self.client.delete(url)

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)



from __future__ import annotations

import hashlib
import importlib
import random

from django.conf import settings


class UserNameGenerator:
    ADJECTIVES = [
        "Quick",
        "Lucky",
        "Happy",
        "Sad",
        "Bright",
        "Dark",
        "Shiny",
        "Gloomy",
        "Cool",
        "Mysterious",
        "Witty",
        "Clever",
        "Brave",
        "Sly",
        "Fierce",
        "Whimsical",
        "Radiant",
        "Spirited",
        "Dynamic",
        "Vibrant",
        "Eloquent",
        "Adventurous",
        "Enigmatic",
        "Serene",
        "Resilient",
        "Tranquil",
        "Majestic",
        "Ethereal",
        "Gallant",
        "Vivacious",
        "Melodic",
        "Zesty",
        "Inquisitive",
        "Dazzling",
        "Empowered",
        "Cheerful",
        "Genuine",
        "Effervescent",
        "Blissful",
        "Savage",
        "Astute",
        "Nimble",
        "Crafty",
        "Audacious",
        "Surreal",
    ]
    NOUNS = [
        "Bear",
        "Fox",
        "Rabbit",
        "Wolf",
        "Hawk",
        "Falcon",
        "Lion",
        "Tiger",
        "Wizard",
        "Dragon",
        "Sorcerer",
        "Knight",
        "Elf",
        "Dwarf",
        "Phoenix",
        "Griffin",
        "Centaur",
        "Mermaid",
        "Siren",
        "Pegasus",
        "Ogre",
        "Goblin",
        "Unicorn",
        "Phoenix",
        "Yeti",
        "Basilisk",
        "Minotaur",
        "Satyr",
        "Chimera",
        "Juggernaut",
        "Mystic",
        "Warlock",
        "Ninja",
        "Samurai",
        "Valkyrie",
        "Sentinel",
        "Behemoth",
        "Enchantress",
        "Oracle",
        "Specter",
        "Corsair",
        "Pirate",
        "Nomad",
        "Sorceress",
        "Voyager",
        "Templar",
        "Rebel",
    ]

    @staticmethod
    def generate_hash(email: str):
        email_bytes = email.encode("utf-8")
        hash_obj = hashlib.shake_256(email_bytes)
        return hash_obj.hexdigest(4)

    @staticmethod
    def generate_username(
        email: str,
        max_length: int = settings.ACCOUNT_USERNAME_MAX_LENGTH,
        max_attempts: int = 1000,
    ):
        user_account_model = importlib.import_module(
            "user.models.account"
        ).UserAccount

        for _attempt in range(max_attempts):
            adjective = random.choice(UserNameGenerator.ADJECTIVES)
            noun = random.choice(UserNameGenerator.NOUNS)
            email_hash = UserNameGenerator.generate_hash(email)
            prefix = f"{adjective}{noun}"
            username = f"{prefix}#{email_hash}"
            if (
                len(username) <= max_length
                and not user_account_model.objects.filter(
                    username=username
                ).exists()
            ):
                return username
        raise RuntimeError(
            "Failed to generate a unique username within the maximum attempts."
        )



from django.contrib.contenttypes.models import ContentType
from django.db import models
from parler.managers import TranslatableManager


class TagManager(TranslatableManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related()
            .prefetch_related("translations")
        )

    def active(self):
        return self.get_queryset().filter(active=True)

    def inactive(self):
        return self.get_queryset().filter(active=False)


class TaggedItemManager(models.Manager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related("tag", "content_type")
            .prefetch_related("tag__translations")
        )

    def active_tags(self):
        return self.get_queryset().filter(tag__active=True)

    def get_tags_for(self, obj_type, obj_id):
        content_type = ContentType.objects.get_for_model(obj_type)

        return self.get_queryset().filter(
            content_type=content_type, object_id=obj_id
        )



from django.utils.translation import gettext_lazy as _
from djmoney.contrib.django_rest_framework import MoneyField
from drf_spectacular.helpers import lazy_serializer
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers

from cart.models import CartItem
from product.serializers.product import ProductSerializer


class CartItemSerializer(serializers.ModelSerializer[CartItem]):
    cart_id = serializers.SerializerMethodField()
    product = ProductSerializer(read_only=True)
    weight_info = serializers.SerializerMethodField(
        help_text=_("Weight information for shipping calculations")
    )
    price = MoneyField(max_digits=11, decimal_places=2, read_only=True)
    final_price = MoneyField(max_digits=11, decimal_places=2, read_only=True)
    discount_value = MoneyField(max_digits=11, decimal_places=2, read_only=True)
    vat_value = MoneyField(max_digits=11, decimal_places=2, read_only=True)
    total_price = MoneyField(max_digits=11, decimal_places=2, read_only=True)
    total_discount_value = MoneyField(
        max_digits=11, decimal_places=2, read_only=True
    )

    def get_cart_id(self, obj: CartItem) -> int:
        return obj.cart.id

    @extend_schema_field(
        {
            "type": "object",
            "properties": {
                "unit_weight": {"type": "number"},
                "total_weight": {"type": "number"},
                "weight_unit": {"type": "string"},
            },
            "required": ["unit_weight", "total_weight", "weight_unit"],
        }
    )
    def get_weight_info(self, obj: CartItem) -> dict | None:
        product = obj.product
        if product.weight:
            total_weight = product.weight.value * obj.quantity
            return {
                "unit_weight": product.weight.value,
                "total_weight": total_weight,
                "weight_unit": product.weight.unit,
            }
        return None

    class Meta:
        model = CartItem
        fields = (
            "id",
            "cart_id",
            "product",
            "quantity",
            "weight_info",
            "price",
            "final_price",
            "discount_value",
            "price_save_percent",
            "discount_percent",
            "vat_percent",
            "vat_value",
            "total_price",
            "total_discount_value",
            "created_at",
            "updated_at",
            "uuid",
        )
        read_only_fields = (
            "id",
            "cart_id",
            "product",
            "weight_info",
            "price",
            "final_price",
            "discount_value",
            "price_save_percent",
            "discount_percent",
            "vat_percent",
            "vat_value",
            "total_price",
            "total_discount_value",
            "created_at",
            "updated_at",
            "uuid",
        )


class CartItemDetailSerializer(CartItemSerializer):
    recommendations = serializers.SerializerMethodField(
        help_text=_("Related products that might interest the customer")
    )

    @extend_schema_field(
        lazy_serializer("product.serializers.product.ProductSerializer")(
            many=True
        )
    )
    def get_recommendations(self, obj: CartItem):
        if obj.product.category:
            related_products = (
                obj.product.category.products.filter(active=True)
                .exclude(id=obj.product.id)
                .order_by("-view_count")[:3]
            )

            return ProductSerializer(
                related_products, many=True, context=self.context
            ).data
        return []

    class Meta(CartItemSerializer.Meta):
        fields = (
            *CartItemSerializer.Meta.fields,
            "recommendations",
        )


class CartItemWriteSerializer(serializers.ModelSerializer[CartItem]):
    def validate_quantity(self, value: int) -> int:
        if value <= 0:
            raise serializers.ValidationError(
                _("Quantity must be greater than zero.")
            )
        if value > 999:
            raise serializers.ValidationError(_("Quantity cannot exceed 999."))
        return value

    def validate(self, data: dict) -> dict:
        if self.instance:
            product = self.instance.product
            quantity = data.get("quantity", self.instance.quantity)
        else:
            product = data.get("product")
            quantity = data.get("quantity", 1)

        if product:
            if not product.active:
                raise serializers.ValidationError(
                    _("Product '{product_name}' is not available.").format(
                        product_name=product.safe_translation_getter(
                            "name", any_language=True
                        ),
                    )
                )

            if product.stock < quantity:
                raise serializers.ValidationError(
                    _(
                        "Not enough stock for '{product_name}'. Available: {product_stock}, Requested: {quantity}"
                    ).format(
                        product_name=product.safe_translation_getter(
                            "name", any_language=True
                        ),
                        product_stock=product.stock,
                        quantity=quantity,
                    )
                )

        return data

    def create(self, validated_data: dict) -> CartItem:
        cart = self.context.get("cart") or validated_data.get("cart")
        product = validated_data.get("product")
        quantity = validated_data.get("quantity")

        if not cart:
            raise serializers.ValidationError(_("Cart is not provided."))

        existing_item = CartItem.objects.filter(
            cart=cart, product=product
        ).first()

        if existing_item:
            new_quantity = existing_item.quantity + quantity
            if product.stock < new_quantity:
                raise serializers.ValidationError(
                    _(
                        "Not enough stock to add {quantity} more. Current in cart: {existing_item_quantity}, Available: {product_stock}"
                    ).format(
                        quantity=quantity,
                        existing_item_quantity=existing_item.quantity,
                        product_stock=product.stock,
                    )
                )
            existing_item.quantity = new_quantity
            existing_item.save()
            return existing_item
        else:
            return CartItem.objects.create(
                cart=cart, product=product, quantity=quantity
            )

    class Meta:
        model = CartItem
        fields = ("id", "cart", "product", "quantity")
        read_only_fields = ("id",)


class CartItemCreateSerializer(serializers.ModelSerializer[CartItem]):
    def validate_quantity(self, value: int) -> int:
        if value <= 0:
            raise serializers.ValidationError(
                _("Quantity must be greater than zero.")
            )
        if value > 999:
            raise serializers.ValidationError(_("Quantity cannot exceed 999."))
        return value

    def validate(self, data: dict) -> dict:
        product = data.get("product")
        quantity = data.get("quantity", 1)

        if product:
            if not product.active:
                raise serializers.ValidationError(
                    _("Product '{product_name}' is not available.").format(
                        product_name=product.safe_translation_getter(
                            "name", any_language=True
                        ),
                    )
                )

            if product.stock < quantity:
                raise serializers.ValidationError(
                    _(
                        "Not enough stock for '{product_name}'. Available: {product_stock}, Requested: {quantity}"
                    ).format(
                        product_name=product.safe_translation_getter(
                            "name", any_language=True
                        ),
                        product_stock=product.stock,
                        quantity=quantity,
                    )
                )
        return data

    def create(self, validated_data: dict) -> CartItem:
        cart = self.context.get("cart") or validated_data.get("cart")
        product = validated_data.get("product")
        quantity = validated_data.get("quantity")

        if not cart:
            raise serializers.ValidationError(_("Cart is not provided."))

        existing_item = CartItem.objects.filter(
            cart=cart, product=product
        ).first()

        if existing_item:
            new_quantity = existing_item.quantity + quantity
            if product.stock < new_quantity:
                raise serializers.ValidationError(
                    _(
                        "Not enough stock to add {quantity} more. Current in cart: {existing_item_quantity}, Available: {product_stock}"
                    ).format(
                        quantity=quantity,
                        existing_item_quantity=existing_item.quantity,
                        product_stock=product.stock,
                    )
                )
            existing_item.quantity = new_quantity
            existing_item.save()
            return existing_item
        else:
            return CartItem.objects.create(
                cart=cart, product=product, quantity=quantity
            )

    class Meta:
        model = CartItem
        fields = ("product", "quantity")


class CartItemUpdateSerializer(serializers.ModelSerializer[CartItem]):
    def validate_quantity(self, value: int) -> int:
        if value <= 0:
            raise serializers.ValidationError(
                _("Quantity must be greater than zero.")
            )
        if value > 999:
            raise serializers.ValidationError(_("Quantity cannot exceed 999."))
        return value

    def validate(self, data: dict) -> dict:
        instance = self.instance
        if instance:
            product = instance.product
            quantity = data.get("quantity", instance.quantity)

            if product and product.stock < quantity:
                raise serializers.ValidationError(
                    _(
                        "Not enough stock for '{product_name}'. Available: {product_stock}, Requested: {quantity}"
                    ).format(
                        product_name=product.safe_translation_getter(
                            "name", any_language=True
                        ),
                        product_stock=product.stock,
                        quantity=quantity,
                    )
                )
        return data

    def update(self, instance: CartItem, validated_data: dict) -> CartItem:
        quantity = validated_data.get("quantity", instance.quantity)
        instance.quantity = quantity
        instance.save()
        return instance

    class Meta:
        model = CartItem
        fields = ("quantity",)



import factory
from django.apps import apps
from django.conf import settings

from notification.enum import NotificationKindEnum
from notification.models.notification import Notification

available_languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]


class NotificationTranslationFactory(factory.django.DjangoModelFactory):
    language_code = factory.Iterator(available_languages)
    title = factory.Faker("sentence", nb_words=6)
    message = factory.Faker("paragraph")
    master = factory.SubFactory(
        "notification.factories.notification.NotificationFactory"
    )

    class Meta:
        model = apps.get_model("notification", "NotificationTranslation")
        django_get_or_create = ("language_code", "master")


class NotificationFactory(factory.django.DjangoModelFactory):
    link = factory.Faker("url")
    kind = factory.Iterator(NotificationKindEnum, getter=lambda c: c.value)

    class Meta:
        model = Notification
        skip_postgeneration_save = True

    @factory.post_generation
    def translations(self, create, extracted, **kwargs):
        if not create:
            return

        translations = extracted or [
            NotificationTranslationFactory(language_code=lang, master=self)
            for lang in available_languages
        ]

        for translation in translations:
            translation.master = self
            translation.save()



# Generated by Django 5.2.1 on 2025-05-14 20:33

import django.contrib.postgres.indexes
from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('blog', '0024_alter_blogauthor_website_and_more'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RenameIndex(
            model_name='blogauthor',
            new_name='blogauthor_created_at_ix',
            old_name='blogauthor_created_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogauthor',
            new_name='blogauthor_updated_at_ix',
            old_name='blogauthor_updated_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogcategory',
            new_name='blogcategory_created_at_ix',
            old_name='blogcategory_created_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogcategory',
            new_name='blogcategory_updated_at_ix',
            old_name='blogcategory_updated_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogcategory',
            new_name='blogcategory_sort_order_ix',
            old_name='blogcategory_sort_order_idx',
        ),
        migrations.RenameIndex(
            model_name='blogcomment',
            new_name='blogcomment_created_at_ix',
            old_name='blogcomment_created_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogcomment',
            new_name='blogcomment_updated_at_ix',
            old_name='blogcomment_updated_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blogpost_created_at_ix',
            old_name='blogpost_created_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blogpost_updated_at_ix',
            old_name='blogpost_updated_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blogpost_published_at_ix',
            old_name='blogpost_published_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blogpost_is_published_ix',
            old_name='blogpost_is_published_idx',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blog_post_view_count_ix',
            old_name='blog_blogpo_view_co_226064_btree',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blog_post_status_ix',
            old_name='blog_blogpo_status_e66397_btree',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blog_post_featured_ix',
            old_name='blog_blogpo_feature_b84efc_btree',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blog_post_category_ix',
            old_name='blog_blogpo_categor_4d71d3_btree',
        ),
        migrations.RenameIndex(
            model_name='blogpost',
            new_name='blog_post_author_ix',
            old_name='blog_blogpo_author__753b9d_btree',
        ),
        migrations.RenameIndex(
            model_name='blogtag',
            new_name='blogtag_created_at_ix',
            old_name='blogtag_created_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogtag',
            new_name='blogtag_updated_at_ix',
            old_name='blogtag_updated_at_idx',
        ),
        migrations.RenameIndex(
            model_name='blogtag',
            new_name='blogtag_sort_order_ix',
            old_name='blogtag_sort_order_idx',
        ),
        migrations.AddIndex(
            model_name='blogauthor',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['user'], name='blog_author_user_ix'),
        ),
        migrations.AddIndex(
            model_name='blogcategory',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['slug'], name='blog_category_slug_ix'),
        ),
        migrations.AddIndex(
            model_name='blogcategory',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['parent'], name='blog_category_parent_ix'),
        ),
        migrations.AddIndex(
            model_name='blogcomment',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['is_approved'], name='blog_blogco_is_appr_b3338c_btree'),
        ),
        migrations.AddIndex(
            model_name='blogcomment',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['post'], name='blog_blogco_post_id_8fff6c_btree'),
        ),
        migrations.AddIndex(
            model_name='blogcomment',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['user'], name='blog_blogco_user_id_78a803_btree'),
        ),
        migrations.AddIndex(
            model_name='blogcomment',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['parent'], name='blog_blogco_parent__6c614f_btree'),
        ),
        migrations.AddIndex(
            model_name='blogpost',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['slug'], name='blog_post_slug_ix'),
        ),
        migrations.AddIndex(
            model_name='blogpost',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['status', 'featured'], name='blog_post_status_featured_ix'),
        ),
    ]



# Generated by Django 5.1.4 on 2024-12-31 00:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('product', '0020_alter_producttranslation_master'),
    ]

    operations = [
        migrations.AlterField(
            model_name='historicalproduct',
            name='seo_description',
            field=models.TextField(blank=True, default='', max_length=300, verbose_name='Seo Description'),
        ),
        migrations.AlterField(
            model_name='historicalproduct',
            name='seo_keywords',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Seo Keywords'),
        ),
        migrations.AlterField(
            model_name='historicalproduct',
            name='seo_title',
            field=models.CharField(blank=True, default='', max_length=70, verbose_name='Seo Title'),
        ),
        migrations.AlterField(
            model_name='product',
            name='seo_description',
            field=models.TextField(blank=True, default='', max_length=300, verbose_name='Seo Description'),
        ),
        migrations.AlterField(
            model_name='product',
            name='seo_keywords',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Seo Keywords'),
        ),
        migrations.AlterField(
            model_name='product',
            name='seo_title',
            field=models.CharField(blank=True, default='', max_length=70, verbose_name='Seo Title'),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='seo_description',
            field=models.TextField(blank=True, default='', max_length=300, verbose_name='Seo Description'),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='seo_keywords',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Seo Keywords'),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='seo_title',
            field=models.CharField(blank=True, default='', max_length=70, verbose_name='Seo Title'),
        ),
        migrations.AlterField(
            model_name='producttranslation',
            name='name',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Name'),
        ),
    ]



from django.db import models


class ContactQuerySet(models.QuerySet):
    def by_email_domain(self, domain):
        return self.filter(email__iendswith=f"@{domain}")


class ContactManager(models.Manager):
    def get_queryset(self) -> ContactQuerySet:
        return ContactQuerySet(self.model, using=self._db)

    def by_email_domain(self, domain):
        return self.get_queryset().by_email_domain(domain)



from django.utils.translation import gettext_lazy as _
from django_filters import rest_framework as filters

from core.filters.camel_case_filters import CamelCaseFilterMixin
from core.filters.core import TimeStampFilterMixin, UUIDFilterMixin
from user.models.account import UserAccount


class UserAccountFilter(
    TimeStampFilterMixin,
    UUIDFilterMixin,
    CamelCaseFilterMixin,
    filters.FilterSet,
):
    id = filters.NumberFilter(
        field_name="id",
        lookup_expr="exact",
        help_text=_("Filter by user account ID"),
    )
    email = filters.CharFilter(
        field_name="email",
        lookup_expr="icontains",
        help_text=_("Filter by email (partial match)"),
    )
    email_exact = filters.CharFilter(
        field_name="email",
        lookup_expr="exact",
        help_text=_("Filter by exact email"),
    )
    username = filters.CharFilter(
        field_name="username",
        lookup_expr="icontains",
        help_text=_("Filter by username (partial match)"),
    )
    username_exact = filters.CharFilter(
        field_name="username",
        lookup_expr="exact",
        help_text=_("Filter by exact username"),
    )
    first_name = filters.CharFilter(
        field_name="first_name",
        lookup_expr="icontains",
        help_text=_("Filter by first name (partial match)"),
    )
    last_name = filters.CharFilter(
        field_name="last_name",
        lookup_expr="icontains",
        help_text=_("Filter by last name (partial match)"),
    )
    is_active = filters.BooleanFilter(
        field_name="is_active",
        help_text=_("Filter by active status"),
    )
    is_staff = filters.BooleanFilter(
        field_name="is_staff",
        help_text=_("Filter by staff status"),
    )
    country = filters.CharFilter(
        field_name="country__alpha_2",
        lookup_expr="exact",
        help_text=_("Filter by country alpha-2 code"),
    )
    country_name = filters.CharFilter(
        field_name="country__translations__name",
        lookup_expr="icontains",
        help_text=_("Filter by country name (partial match)"),
    )
    region = filters.CharFilter(
        field_name="region__alpha",
        lookup_expr="exact",
        help_text=_("Filter by region alpha code"),
    )
    city = filters.CharFilter(
        field_name="city",
        lookup_expr="icontains",
        help_text=_("Filter by city (partial match)"),
    )
    zipcode = filters.CharFilter(
        field_name="zipcode",
        lookup_expr="icontains",
        help_text=_("Filter by zipcode (partial match)"),
    )
    has_phone = filters.BooleanFilter(
        method="filter_has_phone",
        help_text=_("Filter users who have a phone number"),
    )
    has_image = filters.BooleanFilter(
        method="filter_has_image",
        help_text=_("Filter users who have a profile image"),
    )
    has_birth_date = filters.BooleanFilter(
        method="filter_has_birth_date",
        help_text=_("Filter users who have a birth date"),
    )
    has_social_links = filters.BooleanFilter(
        method="filter_has_social_links",
        help_text=_("Filter users who have social media links"),
    )
    full_name = filters.CharFilter(
        method="filter_full_name",
        help_text=_("Filter by full name (first + last name)"),
    )

    def filter_has_phone(self, queryset, name, value):
        if value:
            return queryset.exclude(phone__isnull=True).exclude(phone="")
        else:
            return queryset.filter(phone__isnull=True) | queryset.filter(
                phone=""
            )

    def filter_has_image(self, queryset, name, value):
        if value:
            return queryset.exclude(image__isnull=True).exclude(image="")
        else:
            return queryset.filter(image__isnull=True) | queryset.filter(
                image=""
            )

    def filter_has_birth_date(self, queryset, name, value):
        if value:
            return queryset.exclude(birth_date__isnull=True)
        else:
            return queryset.filter(birth_date__isnull=True)

    def filter_has_social_links(self, queryset, name, value):
        from django.db.models import Q

        social_fields = (
            Q(twitter__gt="")
            | Q(linkedin__gt="")
            | Q(facebook__gt="")
            | Q(instagram__gt="")
            | Q(website__gt="")
            | Q(youtube__gt="")
            | Q(github__gt="")
        )

        if value:
            return queryset.filter(social_fields)
        else:
            return queryset.exclude(social_fields)

    def filter_full_name(self, queryset, name, value):
        from django.contrib.postgres.search import SearchVector

        if value:
            return queryset.annotate(
                full_name_search=SearchVector("first_name", "last_name")
            ).filter(full_name_search=value)
        return queryset

    class Meta:
        model = UserAccount
        fields = {
            "id": ["exact"],
            "email": ["exact", "icontains"],
            "username": ["exact", "icontains"],
            "first_name": ["icontains"],
            "last_name": ["icontains"],
            "is_active": ["exact"],
            "is_staff": ["exact"],
            "country": ["exact"],
            "region": ["exact"],
            "city": ["icontains"],
            "zipcode": ["icontains"],
            "created_at": ["gte", "lte"],
            "updated_at": ["gte", "lte"],
            "uuid": ["exact"],
        }



from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from phonenumber_field.serializerfields import PhoneNumberField
from rest_framework import serializers
from rest_framework.relations import PrimaryKeyRelatedField

from country.models import Country
from region.models import Region
from user.models.address import UserAddress

User = get_user_model()


class UserAddressSerializer(serializers.ModelSerializer[UserAddress]):
    user = PrimaryKeyRelatedField(read_only=True)
    country = PrimaryKeyRelatedField(queryset=Country.objects.all())
    region = PrimaryKeyRelatedField(queryset=Region.objects.all())
    phone = PhoneNumberField()
    mobile_phone = PhoneNumberField(required=False, allow_blank=True)

    class Meta:
        model = UserAddress
        fields = (
            "id",
            "title",
            "first_name",
            "last_name",
            "street",
            "street_number",
            "city",
            "zipcode",
            "floor",
            "location_type",
            "phone",
            "mobile_phone",
            "notes",
            "is_main",
            "user",
            "country",
            "region",
            "created_at",
            "updated_at",
            "uuid",
        )
        read_only_fields = (
            "id",
            "user",
            "created_at",
            "updated_at",
            "uuid",
        )


class UserAddressDetailSerializer(UserAddressSerializer):
    class Meta(UserAddressSerializer.Meta):
        fields = (*UserAddressSerializer.Meta.fields,)


class UserAddressWriteSerializer(serializers.ModelSerializer[UserAddress]):
    user = PrimaryKeyRelatedField(queryset=User.objects.all())
    country = PrimaryKeyRelatedField(queryset=Country.objects.all())
    region = PrimaryKeyRelatedField(
        queryset=Region.objects.all(), required=False
    )
    phone = PhoneNumberField()
    mobile_phone = PhoneNumberField(required=False, allow_blank=True)

    def validate_phone(self, value: str) -> str:
        if not value and not self.initial_data.get("mobile_phone"):
            raise serializers.ValidationError(
                _("Either phone or mobile phone is required")
            )
        return value

    def validate(self, data):
        if data.get("is_main"):
            user = data.get("user") or (
                self.instance.user if self.instance else None
            )
            if user:
                existing_main = UserAddress.objects.filter(
                    user=user, is_main=True
                )
                if self.instance:
                    existing_main = existing_main.exclude(pk=self.instance.pk)

                if existing_main.exists():
                    raise serializers.ValidationError(
                        _("A main address already exists for this user")
                    )

        return data

    class Meta:
        model = UserAddress
        fields = (
            "title",
            "first_name",
            "last_name",
            "street",
            "street_number",
            "city",
            "zipcode",
            "floor",
            "location_type",
            "phone",
            "mobile_phone",
            "notes",
            "is_main",
            "user",
            "country",
            "region",
        )


class ValidateAddressResponseSerializer(serializers.Serializer):
    valid = serializers.BooleanField()
    errors = serializers.DictField(child=serializers.CharField())
    suggestions = serializers.ListField(
        child=serializers.DictField(child=serializers.CharField())
    )


class BulkDeleteAddressesRequestSerializer(serializers.Serializer):
    address_ids = serializers.ListField(
        child=serializers.IntegerField(),
        allow_empty=False,
        help_text=_("List of address IDs to delete"),
    )


class BulkDeleteAddressesResponseSerializer(serializers.Serializer):
    deleted_count = serializers.IntegerField()
    deleted_ids = serializers.ListField(child=serializers.IntegerField())



# Generated by Django 5.1.3 on 2024-11-27 19:34
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("cart", "0006_alter_cartitem_options"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="cart",
            name="cart_created_at_idx",
        ),
        migrations.RemoveIndex(
            model_name="cart",
            name="cart_updated_at_idx",
        ),
        migrations.RemoveIndex(
            model_name="cart",
            name="cart_cart_user_id_0a6acd_btree",
        ),
        migrations.RemoveIndex(
            model_name="cartitem",
            name="cartitem_created_at_idx",
        ),
        migrations.RemoveIndex(
            model_name="cartitem",
            name="cartitem_updated_at_idx",
        ),
        migrations.RemoveIndex(
            model_name="cartitem",
            name="cart_cartit_cart_id_cabd5a_btree",
        ),
    ]



from pathlib import Path
from unittest.mock import patch

from django.test import TestCase

from settings import load_dotenv_file


class TestConfigBase(TestCase):
    @patch("os.path.isfile")
    @patch("dotenv.load_dotenv")
    def test_loads_dotenv_file_when_file_exists(
        self, mock_load_dotenv, mock_isfile
    ):
        dotenv_file = Path(".env").resolve()
        mock_isfile.return_value = True

        load_dotenv_file()

        mock_isfile.assert_called_once_with(dotenv_file)
        mock_load_dotenv.assert_called_once_with(dotenv_file)



from mptt.managers import TreeManager
from mptt.querysets import TreeQuerySet
from parler.managers import TranslatableManager, TranslatableQuerySet


class BlogCategoryQuerySet(TranslatableQuerySet, TreeQuerySet):
    @classmethod
    def as_manager(cls):
        manager = BlogCategoryManager.from_queryset(cls)()
        manager._built_with_as_manager = True
        return manager

    as_manager.queryset_only = True  # type: ignore[attr-defined]


class BlogCategoryManager(TreeManager, TranslatableManager):
    _queryset_class = BlogCategoryQuerySet



from django.conf import settings
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase
from tag.factories.tag import TagFactory
from tag.models.tag import Tag
from core.utils.testing import TestURLFixerMixin

languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]
default_language = settings.PARLER_DEFAULT_LANGUAGE_CODE


class TagViewSetTestCase(TestURLFixerMixin, APITestCase):
    @classmethod
    def setUpTestData(cls):
        cls.tag = TagFactory(active=True)
        cls.inactive_tag = TagFactory(active=False)

    def get_tag_detail_url(self, pk):
        return reverse("tag-detail", args=[pk])

    def get_tag_list_url(self):
        return reverse("tag-list")

    def test_list(self):
        url = self.get_tag_list_url()
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)
        self.assertIn("count", response.data)
        if "links" in response.data:
            self.assertIn("next", response.data["links"])
            self.assertIn("previous", response.data["links"])
        else:
            self.assertIn("next", response.data)
            self.assertIn("previous", response.data)

        self.assertGreaterEqual(len(response.data["results"]), 1)

        first_result = response.data["results"][0]
        expected_basic_fields = {
            "id",
            "translations",
            "active",
            "created_at",
            "updated_at",
            "uuid",
            "sort_order",
            "usage_count",
        }
        self.assertTrue(
            expected_basic_fields.issubset(set(first_result.keys()))
        )

    def test_list_uses_correct_serializer(self):
        url = self.get_tag_list_url()
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        first_result = response.data["results"][0]
        expected_fields = {
            "id",
            "translations",
            "active",
            "created_at",
            "updated_at",
            "uuid",
            "sort_order",
            "usage_count",
        }
        self.assertTrue(expected_fields.issubset(set(first_result.keys())))

    def test_retrieve_valid(self):
        url = self.get_tag_detail_url(self.tag.id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_fields = {
            "id",
            "translations",
            "active",
            "created_at",
            "updated_at",
            "uuid",
            "sort_order",
            "usage_count",
            "content_types",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_retrieve_uses_correct_serializer(self):
        url = self.get_tag_detail_url(self.tag.id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_fields = {
            "id",
            "translations",
            "active",
            "created_at",
            "updated_at",
            "uuid",
            "sort_order",
            "usage_count",
            "content_types",
        }
        response_fields = set(response.data.keys())

        for field in expected_fields:
            self.assertIn(
                field,
                response_fields,
                f"Detail field '{field}' should be present in detail response",
            )

    def test_retrieve_invalid(self):
        invalid_tag_id = 9999
        url = self.get_tag_detail_url(invalid_tag_id)
        response = self.client.get(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_create_valid(self):
        payload = {
            "active": True,
            "translations": {},
        }

        for language in languages:
            language_code = language
            translation_payload = {
                "label": f"New Tag label in {language_code}",
            }
            payload["translations"][language_code] = translation_payload

        url = self.get_tag_list_url()
        response = self.client.post(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        expected_fields = {
            "id",
            "translations",
            "active",
            "created_at",
            "updated_at",
            "uuid",
            "sort_order",
            "usage_count",
            "content_types",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_create_invalid(self):
        payload = {
            "active": "invalid_active",
            "translations": {
                "invalid_lang_code": {
                    "label": "Translation for invalid language code",
                },
            },
        }

        url = self.get_tag_list_url()
        response = self.client.post(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_update_valid(self):
        payload = {
            "active": False,
            "translations": {},
        }

        for language in languages:
            language_code = language
            translation_payload = {
                "label": f"Updated Tag label in {language_code}",
            }
            payload["translations"][language_code] = translation_payload

        url = self.get_tag_detail_url(self.tag.id)
        response = self.client.put(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_fields = {
            "id",
            "translations",
            "active",
            "created_at",
            "updated_at",
            "uuid",
            "sort_order",
            "usage_count",
            "content_types",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_update_invalid(self):
        payload = {
            "active": "invalid_active",
            "translations": {
                "invalid_lang_code": {
                    "label": "Translation for invalid language code",
                },
            },
        }

        url = self.get_tag_detail_url(self.tag.id)
        response = self.client.put(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_partial_update_valid(self):
        payload = {
            "translations": {
                default_language: {"label": "Partially Updated Tag Label"}
            }
        }

        url = self.get_tag_detail_url(self.tag.id)
        response = self.client.patch(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        expected_fields = {
            "id",
            "translations",
            "active",
            "created_at",
            "updated_at",
            "uuid",
            "sort_order",
            "usage_count",
            "content_types",
        }
        self.assertTrue(expected_fields.issubset(set(response.data.keys())))

    def test_partial_update_invalid(self):
        payload = {
            "active": "invalid",
            "translations": {default_language: {"label": "Test Tag"}},
        }

        url = self.get_tag_detail_url(self.tag.id)
        response = self.client.patch(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_destroy_valid(self):
        tag_to_delete = TagFactory(active=True)

        url = self.get_tag_detail_url(tag_to_delete.id)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertFalse(Tag.objects.filter(id=tag_to_delete.id).exists())

    def test_destroy_invalid(self):
        invalid_tag_id = 9999

        url = self.get_tag_detail_url(invalid_tag_id)
        response = self.client.delete(url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_filtering_by_active_status(self):
        url = self.get_tag_list_url()

        response = self.client.get(url, {"active": "true"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for tag_data in response.data["results"]:
            self.assertTrue(tag_data["active"])

        response = self.client.get(url, {"active": "false"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        for tag_data in response.data["results"]:
            self.assertFalse(tag_data["active"])

    def test_filtering_by_id(self):
        url = self.get_tag_list_url()
        response = self.client.get(url, {"id": self.tag.id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(response.data["results"][0]["id"], self.tag.id)

    def test_filtering_by_label(self):
        Tag.objects.all().delete()

        special_tag = TagFactory(active=True)

        from django.apps import apps

        TagTranslation = apps.get_model("tag", "TagTranslation")

        special_tag.translations.all().delete()

        TagTranslation.objects.create(
            master=special_tag,
            language_code=default_language,
            label="SpecialTestTag",
        )

        special_tag.refresh_from_db()

        url = self.get_tag_list_url()
        response = self.client.get(url, {"search": "SpecialTestTag"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        found = False
        for tag_data in response.data["results"]:
            if tag_data["id"] == special_tag.id:
                found = True
                break
        self.assertTrue(found)

    def test_filtering_functionality(self):
        url = self.get_tag_list_url()
        response = self.client.get(url, {"active": "true"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)

    def test_ordering_functionality(self):
        url = self.get_tag_list_url()
        response = self.client.get(url, {"ordering": "created_at"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)

        response = self.client.get(url, {"ordering": "id"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn("results", response.data)

    def test_search_functionality(self):
        url = self.get_tag_list_url()
        response = self.client.get(url, {"search": "test"})
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertIn("results", response.data)

    def test_create_request_response_serializers(self):
        payload = {
            "active": True,
            "translations": {
                default_language: {"label": "Test Tag for Serializers"}
            },
        }

        url = self.get_tag_list_url()
        response = self.client.post(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        expected_detail_fields = {
            "id",
            "translations",
            "active",
            "created_at",
            "updated_at",
            "uuid",
            "sort_order",
            "usage_count",
            "content_types",
        }
        self.assertTrue(
            expected_detail_fields.issubset(set(response.data.keys()))
        )

    def test_validation_errors_consistent(self):
        payload = {
            "active": "not_a_boolean",
            "translations": {},
        }

        url = self.get_tag_list_url()
        response = self.client.post(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertIn("active", response.data)

    def test_sort_order_auto_assignment(self):
        payload = {
            "active": True,
            "sort_order": -1,
            "translations": {default_language: {"label": "Test Tag"}},
        }

        url = self.get_tag_list_url()
        response = self.client.post(url, payload, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertNotEqual(response.data["sort_order"], -1)
        self.assertGreaterEqual(response.data["sort_order"], 0)



from django_filters import rest_framework as filters

from core.filters.camel_case_filters import CamelCaseTimeStampFilterSet
from core.filters.core import SortableFilterMixin, UUIDFilterMixin
from product.models.favourite import ProductFavourite


class ProductFavouriteFilter(
    SortableFilterMixin, UUIDFilterMixin, CamelCaseTimeStampFilterSet
):
    id = filters.NumberFilter(field_name="id")
    product_id = filters.NumberFilter(field_name="product__id")
    user_id = filters.NumberFilter(field_name="user__id")
    uuid = filters.UUIDFilter(field_name="uuid")

    class Meta:
        model = ProductFavourite
        fields = {
            "id": ["exact"],
            "product": ["exact"],
            "user": ["exact"],
            "uuid": ["exact"],
        }



from datetime import date, datetime, timedelta


def date_range(
    start_date: date | datetime,
    end_date: date | datetime,
    step: int = 1,
) -> list[date]:
    if isinstance(start_date, datetime):
        start_date = start_date.date()
    if isinstance(end_date, datetime):
        end_date = end_date.date()

    result = []
    current_date = start_date
    while current_date <= end_date:
        result.append(current_date)
        current_date += timedelta(days=step)

    return result



from __future__ import annotations

from parler.managers import TranslatableManager, TranslatableQuerySet

from product.enum.category import CategoryImageTypeEnum


class CategoryImageQuerySet(TranslatableQuerySet):
    def active(self):
        return self.filter(active=True)

    def by_category(self, category):
        return self.filter(category=category)

    def by_type(self, image_type: CategoryImageTypeEnum):
        return self.filter(image_type=image_type)

    def main_images(self):
        return self.filter(image_type=CategoryImageTypeEnum.MAIN)

    def banner_images(self):
        return self.filter(image_type=CategoryImageTypeEnum.BANNER)

    def icon_images(self):
        return self.filter(image_type=CategoryImageTypeEnum.ICON)

    def gallery_images(self):
        return self.filter(image_type=CategoryImageTypeEnum.GALLERY)

    def get_main_image(self, category):
        return self.filter(
            category=category,
            image_type=CategoryImageTypeEnum.MAIN,
            active=True,
        ).first()

    def get_banner_image(self, category):
        return self.filter(
            category=category,
            image_type=CategoryImageTypeEnum.BANNER,
            active=True,
        ).first()

    def get_icon_image(self, category):
        return self.filter(
            category=category,
            image_type=CategoryImageTypeEnum.ICON,
            active=True,
        ).first()

    def get_image_by_type(self, category, image_type: CategoryImageTypeEnum):
        return self.filter(
            category=category, image_type=image_type, active=True
        ).first()


class CategoryImageManager(TranslatableManager):
    def get_queryset(self) -> CategoryImageQuerySet:
        return CategoryImageQuerySet(self.model, using=self._db)

    def active(self):
        return self.get_queryset().active()

    def by_category(self, category):
        return self.get_queryset().by_category(category)

    def by_type(self, image_type: CategoryImageTypeEnum):
        return self.get_queryset().by_type(image_type)

    def main_images(self):
        return self.get_queryset().main_images()

    def banner_images(self):
        return self.get_queryset().banner_images()

    def icon_images(self):
        return self.get_queryset().icon_images()

    def gallery_images(self):
        return self.get_queryset().gallery_images()

    def get_main_image(self, category):
        return self.get_queryset().get_main_image(category)

    def get_banner_image(self, category):
        return self.get_queryset().get_banner_image(category)

    def get_icon_image(self, category):
        return self.get_queryset().get_icon_image(category)

    def get_image_by_type(self, category, image_type: CategoryImageTypeEnum):
        return self.get_queryset().get_image_by_type(category, image_type)



from django import forms
from django.contrib import messages
from django.core import management
from django.shortcuts import redirect, render
from django.urls import path
from django.utils.translation import gettext_lazy as _
from unfold.sites import UnfoldAdminSite

from core.caches import cache_instance
from core.utils.views import cache_methods_registry


class ClearCacheForm(forms.Form):
    viewset_class = forms.ChoiceField(choices=[])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        choices = [
            (cls.__name__, cls.__name__) for cls in cache_methods_registry
        ]
        self.fields["viewset_class"].choices = choices


class MyAdminSite(UnfoldAdminSite):
    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "clear-cache/",
                self.admin_view(self.clear_cache_view),
                name="clear-cache",
            ),
            path(
                "clear-site-cache/",
                self.admin_view(self.clear_site_cache_view),
                name="clear-site-cache",
            ),
        ]
        return custom_urls + urls

    def clear_cache_view(self, request):
        if request.method == "POST" and "clear_cache_for_class" in request.POST:
            form = ClearCacheForm(request.POST)
            if form.is_valid():
                selected_class = form.cleaned_data["viewset_class"]
                self.clear_cache_for_class(request, selected_class)
                return redirect("admin:clear-cache")
        elif request.method == "POST" and "clear_site_cache" in request.POST:
            self.clear_site_cache()
            messages.success(request, _("Entire site cache cleared"))
            return redirect("admin:clear-cache")
        else:
            form = ClearCacheForm()

        context = {
            **self.each_context(request),
            "form": form,
        }
        return render(request, "admin/clear_cache.html", context)

    @staticmethod
    def clear_cache_for_class(request, class_name):
        cache_keys = cache_instance.keys(class_name)

        if not cache_keys:
            messages.info(request, _("No keys found for %(class_name)s"))
            return

        if cache_keys:
            deleted_keys = 0
            client = cache_instance._cache.get_client()
            for key in cache_keys:
                resp = client.delete(key)
                if resp:
                    deleted_keys += 1

            if deleted_keys > 0:
                messages.success(
                    request,
                    _("Deleted %(deleted_keys)s keys for %(class_name)s"),
                )
            else:
                messages.info(request, _("No keys found for %(class_name)s"))

    def clear_site_cache_view(self, request):
        self.clear_site_cache()
        messages.success(request, _("Entire site cache cleared"))
        return redirect("admin:clear-cache")

    @staticmethod
    def clear_site_cache():
        management.call_command("clear_cache")



from django.template import Context, Template
from django.test import TestCase


class AddClassFilterTest(TestCase):
    def test_add_class_with_existing_class(self):
        template_string = '<div class="existing-class">Content</div>'
        expected_output = (
            '<div class="existing-class custom-class">Content</div>'
        )
        template = Template(
            '{% load form_filters %}{{ value|add_class:"custom-class" }}'
        )
        rendered = template.render(Context({"value": template_string}))
        self.assertEqual(rendered, expected_output)

    def test_add_class_without_existing_class(self):
        template_string = "<div>Content</div>"
        expected_output = '<div class="custom-class">Content</div>'
        template = Template(
            '{% load form_filters %}{{ value|add_class:"custom-class" }}'
        )
        rendered = template.render(Context({"value": template_string}))
        self.assertEqual(rendered, expected_output)

    def test_add_class_with_multiple_existing_classes(self):
        template_string = '<div class="class1 class2">Content</div>'
        expected_output = (
            '<div class="class1 class2 custom-class">Content</div>'
        )
        template = Template(
            '{% load form_filters %}{{ value|add_class:"custom-class" }}'
        )
        rendered = template.render(Context({"value": template_string}))
        self.assertEqual(rendered, expected_output)

    def test_add_class_with_multiple_spaces_between_classes(self):
        template_string = '<div class=" class1     class2   ">Content</div>'
        expected_output = (
            '<div class=" class1     class2    custom-class">Content</div>'
        )
        template = Template(
            '{% load form_filters %}{{ value|add_class:"custom-class" }}'
        )
        rendered = template.render(Context({"value": template_string}))
        self.assertEqual(rendered, expected_output)

    def test_add_class_with_escaped_quotes(self):
        template_string = '<div class="class1 \\"class2\\"">Content</div>'
        expected_output = (
            '<div class="class1 \\"class2\\" custom-class">Content</div>'
        )
        template = Template(
            '{% load form_filters %}{{ value|add_class:"custom-class" }}'
        )
        rendered = template.render(Context({"value": template_string}))
        self.assertEqual(rendered, expected_output)



import re


def sanitize_filename(filename: str):
    filename = re.sub(r"[^A-Za-z0-9_.-]", "", filename)
    filename = re.sub(r"\.+", ".", filename).strip(".")
    filename = filename[:50]
    return filename



# Generated by Django 5.2.3 on 2025-06-21 14:53

import django.db.models.deletion
import parler.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0015_recreate_translatable_history_models'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='orderhistorytranslation',
            options={'default_permissions': (), 'managed': True, 'verbose_name': 'Order History Translation'},
        ),
        migrations.AlterModelOptions(
            name='orderitemhistorytranslation',
            options={'default_permissions': (), 'managed': True, 'verbose_name': 'Order Item History Translation'},
        ),
        migrations.AlterField(
            model_name='orderhistorytranslation',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='orderhistorytranslation',
            name='master',
            field=parler.fields.TranslationsForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='translations', to='order.orderhistory'),
        ),
        migrations.AlterField(
            model_name='orderitemhistorytranslation',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='orderitemhistorytranslation',
            name='master',
            field=parler.fields.TranslationsForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='translations', to='order.orderitemhistory'),
        ),
        migrations.AlterModelTable(
            name='orderhistorytranslation',
            table='order_orderhistory_translation',
        ),
        migrations.AlterModelTable(
            name='orderitemhistorytranslation',
            table='order_orderitemhistory_translation',
        ),
    ]



import io
import os

from django.conf import settings
from django.core.wsgi import get_wsgi_application
from django.utils.functional import SimpleLazyObject

from wsgi.health_check import health_check

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")


def get_allowed_host_lazy():
    return settings.ALLOWED_HOSTS[0]


application = get_wsgi_application()
application = health_check(application, "/health/")

application(
    {
        "REQUEST_METHOD": "GET",
        "SERVER_NAME": SimpleLazyObject(get_allowed_host_lazy),
        "REMOTE_ADDR": "127.0.0.1",
        "SERVER_PORT": 80,
        "PATH_INFO": "/",
        "wsgi.input": io.BytesIO(b""),
        "wsgi.multiprocess": True,
    },
    lambda x, y: None,
)



from django.contrib.auth import get_user_model
from django.urls import reverse
from datetime import timedelta
from django.utils import timezone
from rest_framework.test import APITestCase
from blog.factories.author import BlogAuthorFactory
from blog.factories.category import BlogCategoryFactory
from blog.factories.post import BlogPostFactory
from blog.models.post import BlogPost
from user.factories.account import UserAccountFactory

User = get_user_model()


class BlogPostFilterTest(APITestCase):
    def setUp(self):
        BlogPost.objects.all().delete()

        self.author = BlogAuthorFactory(user=UserAccountFactory())
        self.category = BlogCategoryFactory()

        self.now = timezone.now()

        self.old_post = BlogPostFactory(
            author=self.author,
            category=self.category,
            is_published=True,
            featured=True,
            view_count=1000,
        )
        self.old_post.created_at = self.now - timedelta(days=60)
        self.old_post.published_at = self.now - timedelta(days=59)
        self.old_post.save()

        self.recent_post = BlogPostFactory(
            author=self.author,
            category=self.category,
            is_published=True,
            featured=False,
            view_count=500,
        )
        self.recent_post.created_at = self.now - timedelta(hours=2)
        self.recent_post.published_at = self.now - timedelta(hours=1)
        self.recent_post.save()

        self.draft_post = BlogPostFactory(
            author=self.author,
            category=self.category,
            is_published=False,
            featured=False,
            view_count=0,
        )
        self.draft_post.created_at = self.now - timedelta(hours=3)
        self.draft_post.published_at = None
        self.draft_post.save()

        self.future_post = BlogPostFactory(
            author=self.author,
            category=self.category,
            is_published=True,
            published_at=self.now + timedelta(days=7),
            featured=False,
        )
        self.future_post.created_at = self.now - timedelta(hours=4)
        self.future_post.save()

    def test_timestamp_filters(self):
        url = reverse("blog-post-list")

        created_after_date = self.now - timedelta(days=30)
        response = self.client.get(
            url,
            {"created_after": created_after_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertNotIn(self.old_post.id, result_ids)
        self.assertIn(self.recent_post.id, result_ids)
        self.assertIn(self.draft_post.id, result_ids)
        self.assertIn(self.future_post.id, result_ids)
        self.assertEqual(len(result_ids), 3)

        created_before_date = self.now - timedelta(days=30)
        response = self.client.get(
            url,
            {"created_before": created_before_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.old_post.id, result_ids)

    def test_publishable_filters(self):
        url = reverse("blog-post-list")

        response = self.client.get(url, {"currently_published": "true"})
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.old_post.id, result_ids)
        self.assertIn(self.recent_post.id, result_ids)
        self.assertNotIn(self.draft_post.id, result_ids)
        self.assertNotIn(self.future_post.id, result_ids)

        published_after_date = self.now - timedelta(days=7)
        response = self.client.get(
            url,
            {"published_after": published_after_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.recent_post.id, result_ids)
        self.assertIn(self.future_post.id, result_ids)

    def test_uuid_filter(self):
        url = reverse("blog-post-list")

        response = self.client.get(url, {"uuid": str(self.recent_post.uuid)})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(response.data["results"][0]["id"], self.recent_post.id)

    def test_camel_case_filters(self):
        url = reverse("blog-post-list")

        created_after_date = self.now - timedelta(days=30)
        response = self.client.get(
            url,
            {
                "createdAfter": created_after_date.isoformat(),
                "isPublished": "true",
            },
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.recent_post.id, result_ids)
        self.assertIn(self.future_post.id, result_ids)

        response = self.client.get(
            url,
            {
                "minViewCount": 600,
                "featured": "true",
                "currentlyPublished": "true",
            },
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.old_post.id, result_ids)

    def test_existing_filters_still_work(self):
        url = reverse("blog-post-list")

        response = self.client.get(
            url,
            {"featured": "true", "min_view_count": 500, "is_published": "true"},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.old_post.id, result_ids)

    def test_complex_filter_combinations(self):
        url = reverse("blog-post-list")

        created_after_date = self.now - timedelta(days=90)
        published_before_date = self.now

        response = self.client.get(
            url,
            {
                "createdAfter": created_after_date.isoformat(),
                "publishedBefore": published_before_date.isoformat(),
                "minViewCount": 100,
                "isPublished": "true",
                "ordering": "-viewCount",
            },
        )

        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)

        results = response.data["results"]
        self.assertEqual(results[0]["id"], self.old_post.id)
        self.assertEqual(results[1]["id"], self.recent_post.id)

    def test_filter_with_ordering(self):
        url = reverse("blog-post-list")

        response = self.client.get(
            url, {"currentlyPublished": "true", "ordering": "-createdAt"}
        )
        self.assertEqual(response.status_code, 200)

        results = response.data["results"]
        self.assertEqual(len(results), 2)

        post_ids = [r["id"] for r in results]
        self.assertEqual(post_ids[0], self.recent_post.id)
        self.assertEqual(post_ids[1], self.old_post.id)

    def tearDown(self):
        BlogPost.objects.all().delete()



from django.test import TestCase

from contact.factories import ContactFactory
from contact.models import Contact


class TestContactFactory(TestCase):
    def test_contact_factory_creates_contact(self):
        contact = ContactFactory()

        self.assertIsInstance(contact, Contact)

        self.assertIsNotNone(contact.name)
        self.assertIsNotNone(contact.email)
        self.assertIsNotNone(contact.message)

        self.assertIsNotNone(contact.id)

        retrieved_contact = Contact.objects.get(id=contact.id)
        self.assertEqual(retrieved_contact, contact)

    def test_contact_factory_with_custom_attributes(self):
        custom_name = "Test Name"
        custom_email = "test@example.com"
        custom_message = "Test message"

        contact = ContactFactory(
            name=custom_name,
            email=custom_email,
            message=custom_message,
        )

        self.assertEqual(contact.name, custom_name)
        self.assertEqual(contact.email, custom_email)
        self.assertEqual(contact.message, custom_message)

    def test_contact_factory_get_or_create(self):
        email = "unique@example.com"
        contact1 = ContactFactory(email=email)

        contact2 = ContactFactory(email=email)

        self.assertEqual(contact1.id, contact2.id)

        self.assertEqual(Contact.objects.filter(email=email).count(), 1)

    def test_contact_factory_signal_behavior(self):
        contact = ContactFactory()

        self.assertIsNotNone(contact.id)
        self.assertTrue(Contact.objects.filter(id=contact.id).exists())



from django.contrib.auth import get_user_model
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIClient, APITestCase

from core.enum import FloorChoicesEnum, LocationChoicesEnum
from country.factories import CountryFactory
from order.enum.status import OrderStatus
from pay_way.factories import PayWayFactory
from product.factories.product import ProductFactory
from product.models.product import Product
from region.factories import RegionFactory
from user.factories.account import UserAccountFactory

User = get_user_model()


class CheckoutViewAPITest(APITestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserAccountFactory(num_addresses=0)
        self.country = CountryFactory(num_regions=0)
        self.region = RegionFactory(country=self.country)
        self.pay_way = PayWayFactory()

    def get_checkout_url(self):
        return reverse("order-list")

    def test_successful_order_creation(self):
        self.client.force_authenticate(user=self.user)
        product_1 = ProductFactory.create(
            active=True, stock=20, num_images=0, num_reviews=0
        )
        product_2 = ProductFactory.create(
            active=True, stock=20, num_images=0, num_reviews=0
        )

        order_data = {
            "user": self.user.id,
            "country": self.country.alpha_2,
            "region": self.region.alpha,
            "floor": FloorChoicesEnum.FIRST_FLOOR.value,
            "location_type": LocationChoicesEnum.HOME.value,
            "street": "123 Main St",
            "street_number": "Apt 4B",
            "pay_way": self.pay_way.id,
            "status": OrderStatus.PENDING.value,
            "first_name": "John",
            "last_name": "Doe",
            "email": "test@test.com",
            "zipcode": "12345",
            "city": "Cityville",
            "phone": "2101234567",
            "mobile_phone": "6912345678",
            "customer_notes": "Test notes",
            "shipping_price": "10.00",
            "items": [
                {
                    "product": product_1.id,
                    "quantity": 2,
                },
                {
                    "product": product_2.id,
                    "quantity": 3,
                },
            ],
        }

        url = self.get_checkout_url()
        response = self.client.post(url, data=order_data, format="json")

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        product_1.refresh_from_db()
        product_2.refresh_from_db()

        self.assertEqual(product_1.stock, 20 - 2)
        self.assertEqual(product_2.stock, 20 - 3)

    def test_failed_order_creation(self):
        self.client.force_authenticate(user=self.user)
        product_3 = ProductFactory.create(stock=10, num_images=0, num_reviews=0)
        product_4 = ProductFactory.create(stock=15, num_images=0, num_reviews=0)

        order_data = {
            "user_id": self.user.id,
            "items": [
                {
                    "product": product_3.id,
                    "quantity": 15,
                },
                {"product": product_4.id, "quantity": 3},
            ],
            "pay_way": self.pay_way.id,
        }

        url = self.get_checkout_url()
        response = self.client.post(url, data=order_data, format="json")

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(Product.objects.get(pk=product_3.id).stock, 10)
        self.assertEqual(Product.objects.get(pk=product_4.id).stock, 15)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.conf import settings
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.decorators import action

from rest_framework.response import Response

from core.api.serializers import ErrorResponseSerializer
from core.api.views import BaseModelViewSet
from core.utils.serializers import (
    create_schema_view_config,
    RequestSerializersConfig,
    ResponseSerializersConfig,
)
from core.utils.views import cache_methods
from product.models.category_image import ProductCategoryImage
from product.serializers.category_image import (
    ProductCategoryImageBulkResponseSerializer,
    ProductCategoryImageBulkUpdateSerializer,
    ProductCategoryImageDetailSerializer,
    ProductCategoryImageSerializer,
    ProductCategoryImageWriteSerializer,
)

if TYPE_CHECKING:
    from django.db.models import QuerySet

req_serializers: RequestSerializersConfig = {
    "create": ProductCategoryImageWriteSerializer,
    "update": ProductCategoryImageWriteSerializer,
    "partial_update": ProductCategoryImageWriteSerializer,
    "bulk_update": ProductCategoryImageBulkUpdateSerializer,
}

res_serializers: ResponseSerializersConfig = {
    "create": ProductCategoryImageDetailSerializer,
    "list": ProductCategoryImageSerializer,
    "retrieve": ProductCategoryImageDetailSerializer,
    "update": ProductCategoryImageDetailSerializer,
    "partial_update": ProductCategoryImageDetailSerializer,
    "bulk_update": ProductCategoryImageBulkResponseSerializer,
    "by_category": ProductCategoryImageSerializer,
    "by_type": ProductCategoryImageSerializer,
}

schema_config = create_schema_view_config(
    model_class=ProductCategoryImage,
    display_config={
        "tag": "Product Category Images",
    },
    request_serializers=req_serializers,
    response_serializers=res_serializers,
)


@extend_schema_view(**schema_config)
@cache_methods(settings.DEFAULT_CACHE_TTL, methods=["list", "retrieve"])
class ProductCategoryImageViewSet(BaseModelViewSet):
    queryset = ProductCategoryImage.objects.select_related("category")
    request_serializers = req_serializers
    response_serializers = res_serializers
    filterset_fields = ["id", "category", "image_type", "active"]
    ordering_fields = ["created_at", "image_type", "sort_order"]
    ordering = ["sort_order", "-created_at"]
    search_fields = ["translations__title", "translations__alt_text"]

    def get_queryset(self) -> QuerySet[ProductCategoryImage]:
        queryset = super().get_queryset()

        category_id = self.request.query_params.get("category")
        if category_id:
            queryset = queryset.filter(category_id=category_id)

        image_type = self.request.query_params.get("image_type")
        if image_type:
            queryset = queryset.filter(image_type=image_type)

        active_only = self.request.query_params.get("active_only")
        if active_only and active_only.lower() == "true":
            queryset = queryset.filter(active=True)

        return queryset.distinct()

    @extend_schema(
        operation_id="bulkUpdateProductCategoryImages",
        summary=_("Bulk update category images"),
        description=_(
            "Update multiple category images at once. Can update active status and sort order."
        ),
        tags=["Product Category Images"],
        request=ProductCategoryImageBulkUpdateSerializer,
        responses={
            200: ProductCategoryImageBulkResponseSerializer,
            400: ErrorResponseSerializer,
            401: ErrorResponseSerializer,
        },
    )
    @action(detail=False, methods=["patch"])
    def bulk_update(self, request):
        request_serializer_class = self.get_request_serializer()
        request_serializer = request_serializer_class(data=request.data)
        request_serializer.is_valid(raise_exception=True)

        image_ids = request_serializer.validated_data["image_ids"]
        update_data = {}

        if "active" in request_serializer.validated_data:
            update_data["active"] = request_serializer.validated_data["active"]

        if "sort_order" in request_serializer.validated_data:
            update_data["sort_order"] = request_serializer.validated_data[
                "sort_order"
            ]

        if not update_data:
            return Response(
                {"error": "No fields to update provided."},
                status=status.HTTP_400_BAD_REQUEST,
            )

        updated_count = ProductCategoryImage.objects.filter(
            id__in=image_ids
        ).update(**update_data)

        response_data = {
            "success": True,
            "message": f"Successfully updated {updated_count} category images.",
            "updated_count": updated_count,
        }

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(response_data)
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @extend_schema(
        operation_id="getProductCategoryImagesByCategory",
        summary=_("Get images by category"),
        description=_("Retrieve all images for a specific category."),
        tags=["Product Category Images"],
        responses={
            200: {
                "type": "array",
                "items": {"$ref": "#/components/schemas/ProductCategoryImage"},
            },
            404: ErrorResponseSerializer,
        },
    )
    @action(detail=False, methods=["get"])
    def by_category(self, request):
        category_id = request.query_params.get("category_id")
        if not category_id:
            return Response(
                {"error": "category_id parameter is required."},
                status=status.HTTP_400_BAD_REQUEST,
            )

        try:
            category_id = int(category_id)
        except ValueError:
            return Response(
                {"error": "category_id must be a valid integer."},
                status=status.HTTP_400_BAD_REQUEST,
            )

        images = ProductCategoryImage.objects.filter(category_id=category_id)

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(
            images, many=True, context=self.get_serializer_context()
        )
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @extend_schema(
        operation_id="getProductCategoryImagesByType",
        summary=_("Get images by type"),
        description=_(
            "Retrieve all images of a specific type (main, banner, icon, etc.)."
        ),
        tags=["Product Category Images"],
        responses={
            200: {
                "type": "array",
                "items": {"$ref": "#/components/schemas/ProductCategoryImage"},
            },
        },
    )
    @action(detail=False, methods=["get"])
    def by_type(self, request):
        image_type = request.query_params.get("image_type")
        if not image_type:
            return Response(
                {"error": "image_type parameter is required."},
                status=status.HTTP_400_BAD_REQUEST,
            )

        images = ProductCategoryImage.objects.filter(image_type=image_type)

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(
            images, many=True, context=self.get_serializer_context()
        )
        return Response(response_serializer.data, status=status.HTTP_200_OK)



from measurement.base import BidimensionalMeasure, MeasureBase


def get_measurement(
    measure: type[MeasureBase | BidimensionalMeasure],
    value,
    unit: str | None = None,
    original_unit: str | None = None,
):
    unit = unit or measure.STANDARD_UNIT

    m = measure(**{unit: value})
    if original_unit:
        m.unit = original_unit
    if isinstance(m, BidimensionalMeasure):
        m.reference.value = 1
    return m



from typing import cast

from django.db import models
from django.db.models import ExpressionWrapper, F, QuerySet, Sum
from djmoney.models.fields import MoneyField

from order.enum.status import OrderStatus


class OrderQuerySet(models.QuerySet):
    def with_total_amounts(self) -> QuerySet:
        return self.annotate(
            items_total=Sum(
                ExpressionWrapper(
                    F("items__price") * F("items__quantity"),
                    output_field=MoneyField(max_digits=11, decimal_places=2),
                )
            )
        )

    def pending(self) -> QuerySet:
        return self.filter(status=OrderStatus.PENDING)

    def processing(self) -> QuerySet:
        return self.filter(status=OrderStatus.PROCESSING)

    def shipped(self) -> QuerySet:
        return self.filter(status=OrderStatus.SHIPPED)

    def delivered(self) -> QuerySet:
        return self.filter(status=OrderStatus.DELIVERED)

    def completed(self) -> QuerySet:
        return self.filter(status=OrderStatus.COMPLETED)

    def canceled(self) -> QuerySet:
        return self.filter(status=OrderStatus.CANCELED)

    def returned(self) -> QuerySet:
        return self.filter(status=OrderStatus.RETURNED)

    def refunded(self) -> QuerySet:
        return self.filter(status=OrderStatus.REFUNDED)


class OrderManager(models.Manager):
    def get_queryset(self) -> OrderQuerySet:
        return cast(
            "OrderQuerySet",
            OrderQuerySet(self.model, using=self._db)
            .select_related("user", "pay_way", "country", "region")
            .prefetch_related("items", "items__product")
            .exclude(is_deleted=True),
        )

    def with_total_amounts(self) -> QuerySet:
        return self.get_queryset().with_total_amounts()

    def pending(self) -> QuerySet:
        return self.get_queryset().pending()

    def processing(self) -> QuerySet:
        return self.get_queryset().processing()

    def shipped(self) -> QuerySet:
        return self.get_queryset().shipped()

    def delivered(self) -> QuerySet:
        return self.get_queryset().delivered()

    def completed(self) -> QuerySet:
        return self.get_queryset().completed()

    def canceled(self) -> QuerySet:
        return self.get_queryset().canceled()

    def returned(self) -> QuerySet:
        return self.get_queryset().returned()

    def refunded(self) -> QuerySet:
        return self.get_queryset().refunded()



import uuid

import factory
from django.apps import apps
from django.conf import settings
from faker import Faker

from blog.models.category import BlogCategory
from core.factories import CustomDjangoModelFactory

fake = Faker()

available_languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]


def generate_unique_blog_category_filename():
    return f"blog_category_{uuid.uuid4().hex[:8]}.jpg"


class BlogCategoryTranslationFactory(factory.django.DjangoModelFactory):
    language_code = factory.Iterator(available_languages)
    name = factory.Faker("word")
    description = factory.Faker("paragraph")
    master = factory.SubFactory("blog.factories.category.BlogCategoryFactory")

    class Meta:
        model = apps.get_model("blog", "BlogCategoryTranslation")
        django_get_or_create = ("language_code", "master")


class BlogCategoryFactory(CustomDjangoModelFactory):
    auto_translations = False

    unique_model_fields = [
        ("slug", lambda: fake.slug()),
    ]

    image = factory.django.ImageField(
        filename=factory.LazyFunction(generate_unique_blog_category_filename),
        width=1920,
        height=1080,
        color="green",
        format="JPEG",
    )
    parent = None

    class Meta:
        model = BlogCategory
        django_get_or_create = ("slug",)
        skip_postgeneration_save = True

    @factory.post_generation
    def translations(self, create, extracted, **kwargs):
        if not create:
            return

        translations = extracted or [
            BlogCategoryTranslationFactory(language_code=lang, master=self)
            for lang in available_languages
        ]

        for translation in translations:
            translation.master = self
            translation.save()



from datetime import timedelta

from django.conf import settings
from django.db import models
from django.db.models import Count, F, Sum
from django.utils import timezone
from djmoney.money import Money


class OrderItemQuerySet(models.QuerySet):
    def for_order(self, order):
        return self.filter(order=order)

    def for_product(self, product):
        return self.filter(product=product)

    def with_product_data(self):
        return self.select_related("product").prefetch_related(
            "product__translations"
        )

    def sum_quantities(self):
        return (
            self.aggregate(total_quantity=Sum("quantity"))["total_quantity"]
            or 0
        )

    def annotate_total_price(self):
        return self.annotate(calculated_total=F("price") * F("quantity"))

    def total_items_cost(self):
        items = self.annotate_total_price()
        total = items.aggregate(total=Sum("calculated_total"))["total"] or 0

        first_item = self.first()
        if first_item and hasattr(first_item, "price"):
            return Money(amount=total, currency=first_item.price.currency)
        return Money(amount=0, currency=settings.DEFAULT_CURRENCY)

    def for_user(self, user):
        return self.filter(order__user=user)

    def refunded(self):
        return self.filter(is_refunded=True)

    def not_refunded(self):
        return self.filter(is_refunded=False)

    def recent(self, days=30):
        cutoff_date = timezone.now() - timedelta(days=days)
        return self.filter(order__created_at__gte=cutoff_date)


class OrderItemManager(models.Manager):
    def get_queryset(self) -> OrderItemQuerySet:
        return OrderItemQuerySet(self.model, using=self._db)

    def for_product(self, product):
        return self.get_queryset().for_product(product)

    def for_order(self, order):
        return self.get_queryset().for_order(order)

    def with_product_data(self):
        return self.get_queryset().with_product_data()

    def sum_quantities(self):
        return self.get_queryset().sum_quantities()

    def total_items_cost(self):
        return self.get_queryset().total_items_cost()

    def for_user(self, user):
        return self.get_queryset().for_user(user)

    def refunded(self):
        return self.get_queryset().refunded()

    def not_refunded(self):
        return self.get_queryset().not_refunded()

    def recent(self, days=30):
        return self.get_queryset().recent(days)

    def get_bestselling_products(self, limit=10, days=30):
        return (
            self.recent(days)
            .values("product__id", "product__name")
            .annotate(
                total_quantity=Sum("quantity"),
                total_revenue=Sum(F("price") * F("quantity")),
                order_count=Count("order", distinct=True),
            )
            .order_by("-total_quantity")[:limit]
        )



from .account import UserAccountFactory
from .address import UserAddressFactory
from .subscription import SubscriptionTopicFactory, UserSubscriptionFactory

__all__ = [
    "UserAccountFactory",
    "UserAddressFactory",
    "SubscriptionTopicFactory",
    "UserSubscriptionFactory",
]
